# Copyright (c) 2016 Duke Center for Genomic and Computational Biology
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import errno
import logging
import math
import os
import sys
from argparse import SUPPRESS, ArgumentParser, _ArgumentGroup
import shlex
from typing import Callable, NamedTuple, TypeVar

from toil.batchSystems.abstractBatchSystem import (
    EXIT_STATUS_UNAVAILABLE_VALUE,
    BatchJobExitReason,
    InsufficientSystemResources,
)
from toil.batchSystems.abstractGridEngineBatchSystem import (
    AbstractGridEngineBatchSystem,
)
from toil.batchSystems.options import OptionSetter
from toil.bus import get_job_kind
from toil.common import Config
from toil.job import JobDescription, Requirer
from toil.lib.conversions import strtobool
from toil.lib.misc import CalledProcessErrorStderr, call_command
from toil.statsAndLogging import TRACE

logger = logging.getLogger(__name__)

# We have a complete list of Slurm states. States not in one of these aren't
# allowed. See <https://slurm.schedmd.com/squeue.html#SECTION_JOB-STATE-CODES>

# If a job is in one of these states, Slurm can't run it anymore.
# We don't include states where the job is held or paused here;
# those mean it could run and needs to wait for someone to un-hold
# it, so Toil should wait for it.
#
# We map from each terminal state to the Toil-ontology exit reason.
TERMINAL_STATES: dict[str, BatchJobExitReason] = {
    "BOOT_FAIL": BatchJobExitReason.LOST,
    "CANCELLED": BatchJobExitReason.KILLED,
    "COMPLETED": BatchJobExitReason.FINISHED,
    "DEADLINE": BatchJobExitReason.KILLED,
    "FAILED": BatchJobExitReason.FAILED,
    "NODE_FAIL": BatchJobExitReason.LOST,
    "OUT_OF_MEMORY": BatchJobExitReason.MEMLIMIT,
    "PREEMPTED": BatchJobExitReason.KILLED,
    "REVOKED": BatchJobExitReason.KILLED,
    "SPECIAL_EXIT": BatchJobExitReason.FAILED,
    "TIMEOUT": BatchJobExitReason.KILLED,
}

# If a job is in one of these states, it might eventually move to a different
# state.
NONTERMINAL_STATES: set[str] = {
    "CONFIGURING",
    "COMPLETING",
    "PENDING",
    "RUNNING",
    "RESV_DEL_HOLD",
    "REQUEUE_FED",
    "REQUEUE_HOLD",
    "REQUEUED",
    "RESIZING",
    "SIGNALING",
    "STAGE_OUT",
    "STOPPED",
    "SUSPENDED",
}


def parse_slurm_time(slurm_time: str) -> int:
    """
    Parse a Slurm-style time duration like 7-00:00:00 to a number of seconds.

    Raises ValueError if not parseable.
    """
    # slurm returns time in days-hours:minutes:seconds format
    # Sometimes it will only return minutes:seconds, so days may be omitted
    # For ease of calculating, we'll make sure all the delimeters are ':'
    # Then reverse the list so that we're always counting up from seconds -> minutes -> hours -> days
    total_seconds = 0
    elapsed_split: list[str] = slurm_time.replace("-", ":").split(":")
    elapsed_split.reverse()
    seconds_per_unit = [1, 60, 3600, 86400]
    for index, multiplier in enumerate(seconds_per_unit):
        if index < len(elapsed_split):
            total_seconds += multiplier * int(elapsed_split[index])
    return total_seconds

# For parsing user-provided option overrides (or self-generated
# options) for sbatch, we need a way to recognize long, long-with-equals, and
# short forms.
def option_detector(long: str, short: str | None = None) -> Callable[[str], bool]:
    """
    Get a function that returns true if it sees the long or short
    option.
    """
    def is_match(option: str) -> bool:
        return option == f"--{long}" or option.startswith(f"--{long}=") or (short is not None and option == f"-{short}")
    return is_match

def any_option_detector(options: list[str | tuple[str, str]]) -> Callable[[str], bool]:
    """
    Get a function that returns true if it sees any of the long
    options or long or short option pairs.
    """
    detectors = [option_detector(o) if isinstance(o, str) else option_detector(*o) for o in options]
    def is_match(option: str) -> bool:
        for detector in detectors:
            if detector(option):
                return True
        return False
    return is_match



class SlurmBatchSystem(AbstractGridEngineBatchSystem):
    class PartitionInfo(NamedTuple):
        partition_name: str
        gres: bool
        time_limit: float
        priority: int
        cpus: str
        memory: str

    class PartitionSet:
        """
        Set of available partitions detected on the slurm batch system
        """

        default_gpu_partition: SlurmBatchSystem.PartitionInfo | None
        all_partitions: list[SlurmBatchSystem.PartitionInfo]
        gpu_partitions: set[str]

        def __init__(self) -> None:
            self._get_partition_info()
            self._get_gpu_partitions()

        def _get_gpu_partitions(self) -> None:
            """
            Get all available GPU partitions. Also get the default GPU partition.
            :return: None
            """
            gpu_partitions = [
                partition for partition in self.all_partitions if partition.gres
            ]
            self.gpu_partitions = {p.partition_name for p in gpu_partitions}
            # Grab the lowest priority GPU partition
            # If no GPU partitions are available, then set the default to None
            self.default_gpu_partition = None
            if len(gpu_partitions) > 0:
                self.default_gpu_partition = sorted(
                    gpu_partitions, key=lambda x: x.priority
                )[0]

        def _get_partition_info(self) -> None:
            """
            Call the Slurm batch system with sinfo to grab all available partitions.
            Then parse the output and store all available Slurm partitions
            :return: None
            """
            sinfo_command = ["sinfo", "-a", "-o", "%P %G %l %p %c %m"]

            sinfo = call_command(sinfo_command)

            parsed_partitions = []
            for line in sinfo.split("\n")[1:]:
                if line.strip():
                    partition_name, gres, time, priority, cpus, memory = line.split(" ")
                    try:
                        # Parse time to a number so we can compute on it
                        partition_time: float = parse_slurm_time(time)
                    except ValueError:
                        # Maybe time is unlimited?
                        partition_time = float("inf")
                    try:
                        # Parse priority to an int so we can sort on it
                        partition_priority = int(priority)
                    except ValueError:
                        logger.warning(
                            "Could not parse priority %s for partition %s, assuming high priority",
                            partition_name,
                            priority,
                        )
                        partition_priority = sys.maxsize
                    parsed_partitions.append(
                        SlurmBatchSystem.PartitionInfo(
                            partition_name.rstrip("*"),
                            gres != "(null)",
                            partition_time,
                            partition_priority,
                            cpus,
                            memory,
                        )
                    )
            self.all_partitions = parsed_partitions

        def get_partition(self, time_limit: float | None) -> str | None:
            """
            Get the partition name to use for a job with the given time limit.

            :param time_limit: Time limit in seconds.
            """

            if time_limit is None:
                # Just use Slurm's default
                return None

            winning_partition = None
            for partition in self.all_partitions:
                if partition.time_limit < time_limit:
                    # Can't use this
                    continue
                if winning_partition is None:
                    # Anything beats None
                    winning_partition = partition
                    continue
                if partition.gres and not winning_partition.gres:
                    # Never use a partition witn GRES if you can avoid it
                    continue
                elif not partition.gres and winning_partition.gres:
                    # Never keep a partition with GRES if we find one without
                    winning_partition = partition
                    continue
                if partition.priority > winning_partition.priority:
                    # After that, don't raise priority
                    continue
                elif partition.priority < winning_partition.priority:
                    # And always lower it
                    winning_partition = partition
                    continue
                if partition.time_limit < winning_partition.time_limit:
                    # Finally, lower time limit
                    winning_partition = partition

            # TODO: Store partitions in a better indexed way
            if winning_partition is None and len(self.all_partitions) > 0:
                # We have partitions and none of them can fit this
                raise RuntimeError(
                    f"Could not find a Slurm partition that can fit a job that runs for {time_limit} seconds"
                )

            if winning_partition is None:
                return None
            else:
                return winning_partition.partition_name

    class GridEngineThread(AbstractGridEngineBatchSystem.GridEngineThread):
        # Our boss is always the enclosing class
        boss: SlurmBatchSystem

        def getRunningJobIDs(self) -> dict[int, int]:
            # Should return a dictionary of Job IDs and number of seconds
            times = {}
            with self.runningJobsLock:
                currentjobs: dict[str, int] = {
                    str(self.batchJobIDs[x][0]): x for x in self.runningJobs
                }
            # currentjobs is a dictionary that maps a slurm job id (string) to our own internal job id
            # squeue arguments:
            # -h for no header
            # --format to get jobid i, state %t and time days-hours:minutes:seconds

            lines = call_command(
                ["squeue", "-h", "--format", "%i %t %M"], quiet=True
            ).split("\n")
            for line in lines:
                values = line.split()
                if len(values) < 3:
                    continue
                slurm_jobid, state, elapsed_time = values
                if slurm_jobid in currentjobs and state == "R":
                    try:
                        seconds_running = parse_slurm_time(elapsed_time)
                    except ValueError:
                        # slurm may return INVALID instead of a time
                        seconds_running = 0
                    times[currentjobs[slurm_jobid]] = seconds_running

            return times

        def killJob(self, jobID: int) -> None:
            call_command(["scancel", self.getBatchSystemID(jobID)])

        def prepareSubmission(
            self,
            cpu: int,
            memory: int,
            jobID: int,
            command: str,
            jobName: str,
            job_environment: dict[str, str] | None = None,
            gpus: int | None = None,
        ) -> list[str]:
            # Make sure to use exec so we can get Slurm's signals in the Toil
            # worker instead of having an intervening Bash
            return self.prepareSbatch(
                cpu, memory, jobID, jobName, job_environment, gpus
            ) + [f"--wrap=exec {command}"]

        def submitJob(self, subLine: list[str]) -> int:
            try:
                # Slurm is not quite clever enough to follow the XDG spec on
                # its own. If the submission command sees e.g. XDG_RUNTIME_DIR
                # in our environment, it will send it along (especially with
                # --export=ALL), even though it makes a promise to the job that
                # Slurm isn't going to keep. It also has a tendency to create
                # /run/user/<uid> *at the start* of a job, but *not* keep it
                # around for the duration of the job.
                #
                # So we hide the whole XDG universe from Slurm before we make
                # the submission.
                # Might as well hide DBUS also.
                # This doesn't get us a trustworthy XDG session in Slurm, but
                # it does let us see the one Slurm tries to give us.
                no_session_environment = os.environ.copy()
                session_names = [
                    n
                    for n in no_session_environment.keys()
                    if n.startswith("XDG_") or n.startswith("DBUS_")
                ]
                for name in session_names:
                    del no_session_environment[name]

                output = call_command(subLine, env=no_session_environment)
                # sbatch prints a line like 'Submitted batch job 2954103'
                result = int(output.strip().split()[-1])
                logger.debug("sbatch submitted job %d", result)
                return result
            except OSError as e:
                logger.error(f"sbatch command failed with error: {e}")
                raise e

        def coalesce_job_exit_codes(
            self, batch_job_id_list: list[str]
        ) -> list[int | tuple[int, BatchJobExitReason | None] | None]:
            """
            Collect all job exit codes in a single call.
            :param batch_job_id_list: list of Job ID strings, where each string has the form
            "<job>[.<task>]".
            :return: list of job exit codes or exit code, exit reason pairs associated with the list of job IDs.
            """
            logger.log(
                TRACE, "Getting exit codes for slurm jobs: %s", batch_job_id_list
            )
            # Convert batch_job_id_list to list of integer job IDs.
            job_id_list = [int(id.split(".")[0]) for id in batch_job_id_list]
            status_dict = self._get_job_details(job_id_list)
            exit_codes: list[int | tuple[int, BatchJobExitReason | None] | None] = []
            for _, status in status_dict.items():
                exit_codes.append(self._get_job_return_code(status))
            return exit_codes

        def getJobExitCode(
            self, batchJobID: str
        ) -> int | tuple[int, BatchJobExitReason | None] | None:
            """
            Get job exit code for given batch job ID.
            :param batchJobID: string of the form "<job>[.<task>]".
            :return: integer job exit code.
            """
            logger.log(TRACE, "Getting exit code for slurm job: %s", batchJobID)
            # Convert batchJobID to an integer job ID.
            job_id = int(batchJobID.split(".")[0])
            status_dict = self._get_job_details([job_id])
            status = status_dict[job_id]
            return self._get_job_return_code(status)

        def _get_job_details(
            self, job_id_list: list[int]
        ) -> dict[int, tuple[str | None, int | None]]:
            """
            Helper function for `getJobExitCode` and `coalesce_job_exit_codes`.
            Fetch job details from Slurm's accounting system or job control system.
            :param job_id_list: list of integer Job IDs.
            :return: dict of job statuses, where key is the integer job ID, and value is a tuple
            containing the job's state and exit code.
            """
            try:
                status_dict = self._getJobDetailsFromSacct(job_id_list)
            except (CalledProcessErrorStderr, OSError) as e:
                if isinstance(e, OSError):
                    logger.warning("Could not run sacct: %s", e)
                status_dict = self._getJobDetailsFromScontrol(job_id_list)
            return status_dict

        def _get_job_return_code(
            self, status: tuple[str | None, int | None]
        ) -> int | tuple[int, BatchJobExitReason | None] | None:
            """
            Given a Slurm return code, status pair, summarize them into a Toil return code, exit reason pair.

            The return code may have already been OR'd with the 128-offset
            Slurm-reported signal.

            Slurm will report return codes of 0 even if jobs time out instead
            of succeeding:

                2093597|TIMEOUT|0:0
                2093597.batch|CANCELLED|0:15

            So we guarantee here that, if the Slurm status string is not a
            successful one as defined in
            <https://slurm.schedmd.com/squeue.html#SECTION_JOB-STATE-CODES>, we
            will not return a successful return code.

            Helper function for `getJobExitCode` and `coalesce_job_exit_codes`.
            :param status: tuple containing the job's state and it's return code from Slurm.
            :return: the job's return code for Toil if it's completed, otherwise None.
            """
            state, rc = status

            if state not in TERMINAL_STATES:
                # Don't treat the job as exited yet
                return None

            exit_reason = TERMINAL_STATES[state]

            if exit_reason == BatchJobExitReason.FINISHED:
                # The only state that should produce a 0 ever is COMPLETED. So
                # if the job is COMPLETED and the exit reason is thus FINISHED,
                # pass along the code it has.
                return (rc, exit_reason)  # type: ignore[return-value] # mypy doesn't understand enums well

            if rc == 0:
                # The job claims to be in a state other than COMPLETED, but
                # also to have not encountered a problem. Say the exit status
                # is unavailable.
                return (EXIT_STATUS_UNAVAILABLE_VALUE, exit_reason)

            # If the code is nonzero, pass it along.
            return (rc, exit_reason)  # type: ignore[return-value] # mypy doesn't understand enums well

        def _canonicalize_state(self, state: str) -> str:
            """
            Turn a state string form SLURM into just the state token like "CANCELED".
            """

            # Slurm will sometimes send something like "CANCELED by 30065" in
            # the state column for some reason.

            state_token = state

            if " " in state_token:
                state_token = state.split(" ", 1)[0]

            if (
                state_token not in TERMINAL_STATES
                and state_token not in NONTERMINAL_STATES
            ):
                raise RuntimeError("Toil job in unimplemented Slurm state " + state)

            return state_token

        def _getJobDetailsFromSacct(
            self, job_id_list: list[int]
        ) -> dict[int, tuple[str | None, int | None]]:
            """
            Get SLURM job exit codes for the jobs in `job_id_list` by running `sacct`.
            :param job_id_list: list of integer batch job IDs.
            :return: dict of job statuses, where key is the job-id, and value is a tuple
            containing the job's state and exit code.
            """
            job_ids = ",".join(str(id) for id in job_id_list)
            args = [
                "sacct",
                "-n",  # no header
                "-j",
                job_ids,  # job
                "--format",
                "JobIDRaw,State,ExitCode",  # specify output columns
                "-P",  # separate columns with pipes
                "-S",
                "1970-01-01",
            ]  # override start time limit

            # Collect the job statuses in a dict; key is the job-id, value is a tuple containing
            # job state and exit status. Initialize dict before processing output of `sacct`.
            job_statuses: dict[int, tuple[str | None, int | None]] = {}

            try:
                stdout = call_command(args, quiet=True)
            except OSError as e:
                if e.errno == errno.E2BIG:
                    # Argument list is too big, recurse on half the argument list
                    if len(job_id_list) == 1:
                        # 1 is too big, we can't recurse further, bail out
                        raise
                    job_statuses.update(self._getJobDetailsFromSacct(job_id_list[:len(job_id_list)//2]))
                    job_statuses.update(self._getJobDetailsFromSacct(job_id_list[len(job_id_list)//2:]))
                    return job_statuses
                else:
                    raise

            for job_id in job_id_list:
                job_statuses[job_id] = (None, None)

            for line in stdout.splitlines():
                values = line.strip().split("|")
                if len(values) < 3:
                    continue
                state: str
                job_id_raw, state, exitcode = values
                state = self._canonicalize_state(state)
                logger.log(
                    TRACE, "%s state of job %s is %s", args[0], job_id_raw, state
                )
                # JobIDRaw is in the form JobID[.JobStep]; we're not interested in job steps.
                job_id_parts = job_id_raw.split(".")
                if len(job_id_parts) > 1:
                    continue
                job_id = int(job_id_parts[0])
                status: int
                signal: int
                status, signal = (int(n) for n in exitcode.split(":"))
                if signal > 0:
                    # A non-zero signal may indicate e.g. an out-of-memory killed job
                    status = 128 + signal
                logger.log(
                    TRACE,
                    "%s exit code of job %d is %s, return status %d",
                    args[0],
                    job_id,
                    exitcode,
                    status,
                )
                job_statuses[job_id] = state, status
            logger.log(TRACE, "%s returning job statuses: %s", args[0], job_statuses)
            return job_statuses

        def _getJobDetailsFromScontrol(
            self, job_id_list: list[int]
        ) -> dict[int, tuple[str | None, int | None]]:
            """
            Get SLURM job exit codes for the jobs in `job_id_list` by running `scontrol`.
            :param job_id_list: list of integer batch job IDs.
            :return: dict of job statuses, where key is the job-id, and value is a tuple
            containing the job's state and exit code.
            """
            args = ["scontrol", "show", "job"]
            # `scontrol` can only return information about a single job,
            # or all the jobs it knows about.
            if len(job_id_list) == 1:
                args.append(str(job_id_list[0]))

            stdout = call_command(args, quiet=True)

            # Job records are separated by a blank line.
            job_records = None
            if isinstance(stdout, str):
                job_records = stdout.strip().split("\n\n")
            elif isinstance(stdout, bytes):
                job_records = stdout.decode("utf-8").strip().split("\n\n")

            # Collect the job statuses in a dict; key is the job-id, value is a tuple containing
            # job state and exit status. Initialize dict before processing output of `scontrol`.
            job_statuses: dict[int, tuple[str | None, int | None]] = {}
            job_id: int | None
            for job_id in job_id_list:
                job_statuses[job_id] = (None, None)

            # `scontrol` will report "No jobs in the system", if there are no jobs in the system,
            # and if no job-id was passed as argument to `scontrol`.
            if len(job_records) > 0 and job_records[0] == "No jobs in the system":
                return job_statuses

            for record in job_records:
                job: dict[str, str] = {}
                job_id = None
                for line in record.splitlines():
                    for item in line.split():
                        # Output is in the form of many key=value pairs, multiple pairs on each line
                        # and multiple lines in the output. Each pair is pulled out of each line and
                        # added to a dictionary.
                        # Note: In some cases, the value itself may contain white-space. So, if we find
                        # a key without a value, we consider that key part of the previous value.
                        bits = item.split("=", 1)
                        if len(bits) == 1:
                            job[key] += " " + bits[0]  # type: ignore[has-type]  # we depend on the previous iteration to populate key
                        else:
                            key = bits[0]
                            job[key] = bits[1]
                    # The first line of the record contains the JobId. Stop processing the remainder
                    # of this record, if we're not interested in this job.
                    job_id = int(job["JobId"])
                    if job_id not in job_id_list:
                        logger.log(
                            TRACE, "%s job %d is not in the list", args[0], job_id
                        )
                        break
                if job_id is None or job_id not in job_id_list:
                    continue
                state = job["JobState"]
                state = self._canonicalize_state(state)
                logger.log(TRACE, "%s state of job %s is %s", args[0], job_id, state)
                try:
                    exitcode = job["ExitCode"]
                    if exitcode is not None:
                        status, signal = (int(n) for n in exitcode.split(":"))
                        if signal > 0:
                            # A non-zero signal may indicate e.g. an out-of-memory killed job
                            status = 128 + signal
                        logger.log(
                            TRACE,
                            "%s exit code of job %d is %s, return status %d",
                            args[0],
                            job_id,
                            exitcode,
                            status,
                        )
                        rc = status
                    else:
                        rc = None
                except KeyError:
                    rc = None
                job_statuses[job_id] = (state, rc)
            logger.log(TRACE, "%s returning job statuses: %s", args[0], job_statuses)
            return job_statuses

        ###
        ### Implementation-specific helper methods
        ###

        def prepareSbatch(
            self,
            cpu: int,
            mem: int,
            jobID: int,
            jobName: str,
            job_environment: dict[str, str] | None,
            gpus: int | None,
        ) -> list[str]:
            """
            Returns the sbatch command line to run to queue the job.
            """

            # Start by naming the job
            sbatch_line = ["sbatch", "-J", f"toil_job_{jobID}_{jobName}"]

            # Make sure the job gets a signal before it disappears so that e.g.
            # container cleanup finally blocks can run. Ask for SIGINT so we
            # can get the default Python KeyboardInterrupt which third-party
            # code is likely to plan for. Make sure to send it to the batch
            # shell process with "B:", not to all the srun steps it launches
            # (because there shouldn't be any). We cunningly replaced the batch
            # shell process with the Toil worker process, so Toil should be
            # able to get the signal.
            #
            # TODO: Add a way to detect when the job failed because it
            # responded to this signal and use the right exit reason for it.
            sbatch_line.append("--signal=B:INT@30")

            environment = {}
            environment.update(self.boss.environment)
            if job_environment:
                environment.update(job_environment)

            # "Native extensions" for SLURM (see DRMAA or SAGA)
            # Also any extra arguments from --slurmArgs or TOIL_SLURM_ARGS
            nativeConfig: str = self.boss.config.slurm_args  # type: ignore[attr-defined]

            is_any_mem_option = any_option_detector(["mem", "mem-per-cpu", "mem-per-gpu"])
            is_any_cpus_option = any_option_detector([("cpus-per-task", "c"), "cpus-per-gpu"])
            is_export_option = option_detector("export")
            is_export_file_option = option_detector("export-file")
            is_time_option = option_detector("time", "t")
            is_partition_option = option_detector("partition", "p")

            # We will fill these in with stuff parsed from TOIL_SLURM_ARGS, or
            # with our own determinations if they aren't there.

            # --export=[ALL,]<environment_toil_variables>
            export_all = True
            export_list = [] # Some items here may be multiple comma-separated values
            time_limit: int | None = self.boss.config.slurm_time  # type: ignore[attr-defined]
            partition: str | None = None

            if nativeConfig is not None:
                logger.debug(
                    "Native SLURM options appended to sbatch: %s", nativeConfig
                )

                # Do a mini argument parse to pull out export and parse time if
                # needed
                args = shlex.split(nativeConfig)
                i = 0
                while i < len(args):
                    arg = args[i]
                    if is_any_mem_option(arg) or is_any_cpus_option(arg):
                        # Prohibit arguments that set CPUs or memory
                        raise ValueError(
                            f"Cannot use Slurm argument {arg} which conflicts "
                            f"with Toil's own arguments to Slurm"
                        )
                    elif is_export_option(arg):
                        # Capture the export argument value so we can modify it
                        export_all = False
                        if "=" not in arg:
                            if i + 1 >= len(args):
                                raise ValueError(
                                    f"No value supplied for Slurm {arg} argument"
                                )
                            i += 1
                            export_list.append(args[i])
                        else:
                            export_list.append(arg.split("=", 1)[1])
                    elif is_export_file_option(arg):
                        # Keep --export-file but turn off --export=ALL in that
                        # case.
                        export_all = False
                        sbatch_line.append(arg)
                    elif is_time_option(arg):
                        # Capture the time limit in seconds so we can use it for picking a partition
                        if "=" not in arg:
                            if i + 1 >= len(args):
                                raise ValueError(
                                    f"No value supplied for Slurm {arg} argument"
                                )
                            i += 1
                            time_string = args[i]
                        else:
                            time_string = arg.split("=", 1)[1]
                        time_limit = parse_slurm_time(time_string)
                    elif is_partition_option(arg):
                        # Capture the partition so we can run checks on it and know not to assign one
                        if "=" not in arg:
                            if i + 1 >= len(args):
                                raise ValueError(
                                    f"No value supplied for Slurm {arg} argument"
                                )
                            i += 1
                            partition = args[i]
                        else:
                            partition = arg.split("=", 1)[1]
                    else:
                        # Other arguments pass through.
                        sbatch_line.append(arg)
                    i += 1

            if export_all:
                # We don't have any export overrides so we ened to start with
                # an ALL
                export_list.append("ALL")

            if environment:
                argList = []

                for k, v in environment.items():
                    # TODO: The sbatch man page doesn't say we can quote these;
                    # if we need to send characters like , itself we need to
                    # use --export-file and clean it up when the command has
                    # been issued.
                    quoted_value = shlex.quote(os.environ[k] if v is None else v)
                    argList.append(f"{k}={quoted_value}")

                export_list.extend(argList)

            # If partition isn't set and we have a GPU partition override
            # that applies, apply it
            gpu_partition_override: str | None = self.boss.config.slurm_gpu_partition  # type: ignore[attr-defined]
            if partition is None and gpus and gpu_partition_override:
                partition = gpu_partition_override

            # If partition isn't set and we have a parallel partition override
            # that applies, apply it
            parallel_env: str | None = self.boss.config.slurm_pe  # type: ignore[attr-defined]
            if partition is None and cpu and cpu > 1 and parallel_env:
                partition = parallel_env

            # If partition isn't set and we have a general partition override
            # that applies, apply it
            partition_override: str | None = self.boss.config.slurm_partition  # type: ignore[attr-defined]
            if partition is None and partition_override:
                partition = partition_override

            if partition is None and gpus:
                # Send to a GPU partition
                gpu_partition = self.boss.partitions.default_gpu_partition
                if gpu_partition is None:
                    # no gpu partitions are available, raise an error
                    raise RuntimeError(
                        f"The job {jobName} is requesting GPUs, but the Slurm cluster does not appear to have an accessible partition with GPUs"
                    )
                if (
                    time_limit is not None
                    and gpu_partition.time_limit < time_limit
                ):
                    # TODO: find the lowest-priority GPU partition that has at least each job's time limit!
                    logger.warning(
                        "Trying to submit a job that needs %s seconds to partition %s that has a limit of %s seconds",
                        time_limit,
                        gpu_partition.partition_name,
                        gpu_partition.time_limit,
                    )
                partition = gpu_partition.partition_name

            if partition is None:
                # Pick a partition based on time limit
                partition = self.boss.partitions.get_partition(time_limit)

            # Now generate all the arguments
            if len(export_list) > 0:
                # add --export to the sbatch
                sbatch_line.append("--export=" + ",".join(export_list))
            if partition is not None:
                sbatch_line.append(f"--partition={partition}")
            if gpus:
                # Generate GPU assignment argument
                sbatch_line.append(f"--gres=gpu:{gpus}")
                if partition is not None and partition not in self.boss.partitions.gpu_partitions:
                    # the specified partition is not compatible, so warn the user that the job may not work
                    logger.warning(
                        f"Job {jobName} needs GPUs, but specified partition {partition} does not have them. This job may not work."
                        f"Try specifying one of these partitions instead: {', '.join(self.boss.partitions.gpu_partitions)}."
                    )
            if mem is not None and self.boss.config.slurm_allocate_mem:  # type: ignore[attr-defined]
                # memory passed in is in bytes, but slurm expects megabytes
                sbatch_line.append(f"--mem={math.ceil(mem / 2 ** 20)}")
            if cpu is not None:
                sbatch_line.append(f"--cpus-per-task={math.ceil(cpu)}")
            if time_limit is not None:
                # Put all the seconds in the seconds slot
                sbatch_line.append(f"--time=0:{time_limit}")

            stdoutfile: str = self.boss.format_std_out_err_path(jobID, "%j", "out")
            stderrfile: str = self.boss.format_std_out_err_path(jobID, "%j", "err")
            sbatch_line.extend(["-o", stdoutfile, "-e", stderrfile])

            return sbatch_line

    def __init__(
        self, config: Config, maxCores: float, maxMemory: float, maxDisk: float
    ) -> None:
        super().__init__(config, maxCores, maxMemory, maxDisk)
        self.partitions = SlurmBatchSystem.PartitionSet()

    # Override issuing jobs so we can check if we need to use Slurm's magic
    # whole-node-memory feature.
    def issueBatchJob(
        self,
        command: str,
        job_desc: JobDescription,
        job_environment: dict[str, str] | None = None,
    ) -> int:
        # Avoid submitting internal jobs to the batch queue, handle locally
        local_id = self.handleLocalJob(command, job_desc)
        if local_id is not None:
            return local_id
        else:
            self.check_resource_request(job_desc)
            gpus = self.count_needed_gpus(job_desc)
            job_id = self.getNextJobID()
            self.currentJobs.add(job_id)

            if "memory" not in job_desc.requirements and self.config.slurm_default_all_mem:  # type: ignore[attr-defined]
                # The job doesn't have its own memory requirement, and we are
                # defaulting to whole node memory. Use Slurm's 0-memory sentinel.
                memory = 0
            else:
                # Use the memory actually on the job, or the Toil default memory
                memory = job_desc.memory

            self.newJobsQueue.put(
                (
                    job_id,
                    job_desc.cores,
                    memory,
                    command,
                    get_job_kind(job_desc.get_names()),
                    job_environment,
                    gpus,
                )
            )
            logger.debug(
                "Issued the job command: %s with job id: %s and job name %s",
                command,
                str(job_id),
                get_job_kind(job_desc.get_names()),
            )
        return job_id

    def _check_accelerator_request(self, requirer: Requirer) -> None:
        for accelerator in requirer.accelerators:
            if accelerator["kind"] != "gpu":
                raise InsufficientSystemResources(
                    requirer,
                    "accelerators",
                    details=[
                        f"The accelerator {accelerator} could not be provided"
                        "The Toil Slurm batch system only supports gpu accelerators at the moment."
                    ],
                )

    ###
    ### The interface for SLURM
    ###

    # `scontrol show config` can get us the slurm config, and there are values
    # SchedulerTimeSlice and AcctGatherNodeFreq in there, but
    # SchedulerTimeSlice is for time-sharing preemtion and AcctGatherNodeFreq
    # is for reporting resource statistics (and can be 0). Slurm does not
    # actually seem to have a scheduling granularity or tick rate. So we don't
    # implement getWaitDuration().

    @classmethod
    def add_options(cls, parser: ArgumentParser | _ArgumentGroup) -> None:

        parser.add_argument(
            "--slurmAllocateMem",
            dest="slurm_allocate_mem",
            type=strtobool,
            default=True,
            env_var="TOIL_SLURM_ALLOCATE_MEM",
            help="If False, do not use --mem. Used as a workaround for Slurm clusters that reject jobs "
            "with memory allocations.",
        )
        # Keep these deprcated options for backward compatibility
        parser.add_argument(
            "--dont_allocate_mem",
            action="store_false",
            dest="slurm_allocate_mem",
            help=SUPPRESS,
        )
        parser.add_argument(
            "--allocate_mem",
            action="store_true",
            dest="slurm_allocate_mem",
            help=SUPPRESS,
        )

        parser.add_argument(
            "--slurmDefaultAllMem",
            dest="slurm_default_all_mem",
            type=strtobool,
            default=False,
            env_var="TOIL_SLURM_DEFAULT_ALL_MEM",
            help="If True, assign Toil jobs without their own memory requirements all available "
            "memory on a Slurm node (via Slurm --mem=0).",
        )
        parser.add_argument(
            "--slurmTime",
            dest="slurm_time",
            type=parse_slurm_time,
            default=None,
            env_var="TOIL_SLURM_TIME",
            help="Slurm job time limit, in [DD-]HH:MM:SS format.",
        )
        parser.add_argument(
            "--slurmPartition",
            dest="slurm_partition",
            default=None,
            env_var="TOIL_SLURM_PARTITION",
            help="Partition to send Slurm jobs to.",
        )
        parser.add_argument(
            "--slurmGPUPartition",
            dest="slurm_gpu_partition",
            default=None,
            env_var="TOIL_SLURM_GPU_PARTITION",
            help="Partition to send Slurm jobs to if they ask for GPUs.",
        )
        parser.add_argument(
            "--slurmPE",
            dest="slurm_pe",
            default=None,
            env_var="TOIL_SLURM_PE",
            help="Special partition to send Slurm jobs to if they ask for more than 1 CPU.",
        )
        parser.add_argument(
            "--slurmArgs",
            dest="slurm_args",
            default="",
            env_var="TOIL_SLURM_ARGS",
            help="Extra arguments to pass to Slurm.",
        )

    OptionType = TypeVar("OptionType")

    @classmethod
    def setOptions(cls, setOption: OptionSetter) -> None:
        setOption("slurm_allocate_mem")
        setOption("slurm_default_all_mem")
        setOption("slurm_time")
        setOption("slurm_partition")
        setOption("slurm_gpu_partition")
        setOption("slurm_pe")
        setOption("slurm_args")



#!/usr/bin/env python3

"""
mypy-before-push.py: Git pre-push script to make sure that you know if something you push is going to fail MyPy linting.

Install with:

ln -rs ./contrib/hooks/mypy-before-push.py .git/hooks/pre-push
"""

import sys
import subprocess
import os
from typing import Tuple, Optional

from lib import complain, announce, in_acceptable_environment, write_cache, read_cache, check_to_cache, get_current_commit

def check_can_run(local_object) -> bool:
    """
    Make sure we would be able to run mypy on the given commit.
    """

    if not in_acceptable_environment():
        announce('Environment not set up for type checking.')
        return False
    try:
        current_object = get_current_commit()
    except:
        announce('Currently checked-out commit cannot be determined.')
        return False
    if current_object != local_object:
        announce(f'Commit being pushed is not currently checked out')
        return False
    return True

def check_checked_out_commit(local_object) -> bool:
    """
    If the checked-out commit does not type-check, return false. Else, return true.
    """

    status, log = read_cache(local_object)
    if status is None:
        announce('Commit has not been type-checked. Checking now.')
        status, log, _ = check_to_cache(local_object)
        from_cache = False
    else:
        from_cache = True
    if status == False:
        complain('Commit failed type-checking:')
        sys.stderr.write(log)
        # Type-check again in the background in case something has actually
        # changed about the environment.
        if from_cache:
            if os.fork() == 0:
                check_to_cache(local_object)
                sys.exit(0)
            else:
                announce('Re-checking in the background in case something changed.')
        return False
    elif status == True:
        announce('Commit passed type-checking.')
        return True

def main(argc, argv):
    for line in sys.stdin:
        line = line.strip()
        if line == '':
            continue
        parts = line.split()
        # Pushes will come from standard input as:
        # <local ref> SP <local object name> SP <remote ref> SP <remote object name> LF
        # See <https://www.git-scm.com/docs/githooks#_pre_push>
        local_ref = parts[0]
        local_object = parts[1]
        if local_ref == '(delete)':
            # Deleting a branch. Nothing to do
            continue
        if not check_can_run(local_object):
            announce('Cannot check the commit being pushed.')
            return 0
        if not check_checked_out_commit(local_object):
            complain('You should not push this. CI would fail!')
            return 1
    return 0

if __name__ == "__main__":
    sys.exit(main(len(sys.argv), sys.argv))



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This script is a template for src/toil/version.py. Running it without arguments echoes all
globals, i.e. module attributes. Constant assignments will be echoed verbatim while callables
will be invoked and their result echoed as an assignment using the function name as the left-hand
side and the return value of the function as right-hand side. To prevent a module attribute from
being echoed, start or end the attribute name with an underscore. To print the value of a single
symbol, pass the name of that attribute to the script as a command line argument. You can also
import the expand_ function and invoke it directly with either no or exactly one argument."""

# Note to maintainers:
#
#  - don't import at module level unless you want the imported value to be
#    included in the output
#  - only import from the Python standard run-time library (you can't have any
#    dependencies)
#  - don't import even standard modules at global scope without renaming them
#    to have leading/trailing underscores

baseVersion = "8.2.0"
cgcloudVersion = "1.6.0a1.dev393"


def version():
    """
    A version identifier that includes the full-length commit SHA1 and an optional suffix to
    indicate that the working copy is dirty.
    """
    return "-".join(
        filter(None, [distVersion(), currentCommit(), ("dirty" if dirty() else None)])
    )


def cacheTag():
    """
    A Docker tag that we should use to cache Docker image build layers for this commit.
    """

    import os

    return "".join(
        [
            "cache-",
            # Pick up branch or tag from Gitlagb CI, or just use "local" for everyone.
            (
                (os.getenv("CI_COMMIT_BRANCH", "") + os.getenv("CI_COMMIT_TAG", ""))
                or "local"
            ).replace("/", "-"),
            _pythonVersionSuffix(),
        ]
    )


def mainCacheTag():
    """
    A Docker tag where the Toil mainline builds cache their layers.
    """

    return "".join(["cache-master", _pythonVersionSuffix()])


def distVersion():
    """The distribution version identifying a published release on PyPI."""
    return baseVersion


def exactPython():
    """
    Returns the Python command for the exact version of Python we are installed
    for. Something like 'python2.7' or 'python3.6'.
    """
    import sys

    return f"python{sys.version_info[0]}.{sys.version_info[1]}"


def python():
    """
    Returns the Python command for the general version of Python we are
    installed for.  Something like 'python2.7' or 'python3'.

    We assume all Python 3s are sufficiently intercompatible that we can just
    use 'python3' here for all of them. This is useful because the Toil Docker
    appliance is only built for particular Python versions, and we would like
    workflows to work with a variety of leader Python versions.
    """
    return exactPython()


def _pythonVersionSuffix():
    """
    Returns a short string identifying the running version of Python. Toil
    appliances running the same Toil version but on different versions of
    Python as returned by this function are not compatible.
    """
    import sys

    # For now, we assume all Python 3 releases are intercompatible.
    # We also only tag the Python 2 releases specially, since Python 2 is old and busted.
    return f"-py{sys.version_info[0]}.{sys.version_info[1]}"


def dockerTag():
    """The primary tag of the Docker image for the appliance. This uniquely identifies the appliance image."""
    return version() + _pythonVersionSuffix()


def currentCommit():
    import os
    from subprocess import check_output

    try:
        git_root_dir = os.path.dirname(os.path.abspath(__file__))
        output = (
            check_output(
                f"git log --pretty=oneline -n 1 -- {git_root_dir}",
                shell=True,
                cwd=git_root_dir,
            )
            .decode("utf-8")
            .split()[0]
        )
    except:
        # Return this if we are not in a git environment.
        return "000"
    if isinstance(output, bytes):
        return output.decode("utf-8")
    return str(output)


def dockerRegistry():
    import os

    return os.getenv("TOIL_DOCKER_REGISTRY", "quay.io/ucsc_cgl")


def dockerName():
    import os

    return os.getenv("TOIL_DOCKER_NAME", "toil")


def dirty():
    import os
    from subprocess import call

    try:
        git_root_dir = os.path.dirname(os.path.abspath(__file__))
        return 0 != call(
            "(git diff --exit-code && git diff --cached --exit-code) > /dev/null",
            shell=True,
            cwd=git_root_dir,
        )
    except:
        return False  # In case the git call fails.


def expand_(name=None, others=None):
    """
    Returns a string of all the globals and additional variables passed as the
    others keyword argument.

    :param str name: If set, only the value of the given symbol is returned.
    :param dict others: A dictionary of additional variables to be included in
                        the return value.
    """
    variables = {
        k: v
        for k, v in globals().items()
        if not k.startswith("_") and not k.endswith("_")
    }

    if others is not None:
        variables.update(others)

    def resolve(k):
        v = variables.get(k, None)
        if callable(v):
            v = v()
        return v

    if name is None:
        return "".join(f"{k} = {repr(resolve(k))}\n" for k, v in variables.items())
    else:
        return resolve(name)


def _main():
    import sys

    sys.stdout.write(expand_(*sys.argv[1:]))


if __name__ == "__main__":
    _main()



from _typeshed import Incomplete

from ..core.topicdefnprovider import ITopicDefnProvider
from ..core.topictreetraverser import ITopicTreeVisitor

TOPIC_TREE_FROM_FILE: str

class XmlTopicDefnProvider(ITopicDefnProvider):
    class XmlParserError(RuntimeError): ...
    class UnrecognizedSourceFormatError(ValueError): ...

    def __init__(self, xml, format=...) -> None: ...
    def getDefn(self, topicNameTuple): ...
    def topicNames(self): ...
    def getTreeDoc(self): ...

class XmlVisitor(ITopicTreeVisitor):
    tree: Incomplete
    known_topics: Incomplete
    def __init__(self, elem) -> None: ...

def exportTopicTreeSpecXml(
    moduleName: Incomplete | None = ...,
    rootTopic: Incomplete | None = ...,
    bak: str = ...,
    moduleDoc: Incomplete | None = ...,
): ...



import logging
import multiprocessing
import os
from pathlib import Path
import random
import time
import traceback
from functools import partial

from toil.lib.threading import LastProcessStandingArena, cpu_count, global_mutex

log = logging.getLogger(__name__)


class TestThreading:
    """Test Toil threading/synchronization tools."""

    def testGlobalMutexOrdering(self, tmp_path: Path) -> None:
        for it in range(10):
            log.info("Iteration %d", it)

            scope = tmp_path / f"tempDir{it}"
            scope.mkdir()
            mutex = "mutex"
            # Use processes (as opposed to threads) to prevent GIL from ordering things artificially
            pool = multiprocessing.Pool(processes=cpu_count())
            try:
                numTasks = 100
                temp_results = pool.map_async(
                    func=partial(_testGlobalMutexOrderingTask, scope, mutex),
                    iterable=list(range(numTasks)),
                )
                results = temp_results.get()
            finally:
                pool.close()
                pool.join()

            assert len(results) == numTasks
            for item in results:
                # Make sure all workers say they succeeded
                assert item is True

    def testLastProcessStanding(self, tmp_path: Path) -> None:
        for it in range(10):
            log.info("Iteration %d", it)

            scope = tmp_path / f"tempDir{it}"
            scope.mkdir()
            arena_name = "thunderdome"
            # Use processes (as opposed to threads) to prevent GIL from ordering things artificially
            pool = multiprocessing.Pool(processes=cpu_count())
            try:
                numTasks = 100
                temp_results = pool.map_async(
                    func=partial(_testLastProcessStandingTask, scope, arena_name),
                    iterable=list(range(numTasks)),
                )
                results = temp_results.get()
            finally:
                pool.close()
                pool.join()

            assert len(results) == numTasks
            for item in results:
                # Make sure all workers say they succeeded
                assert item is True
            for filename in os.listdir(scope):
                assert not filename.startswith(
                    "precious"
                ), f"File {filename} still exists"


def _testGlobalMutexOrderingTask(scope: Path, mutex: str, number: int) -> bool:
    try:
        # We will all fight over the potato
        potato = scope / "potato"

        with global_mutex(scope, mutex):
            log.info("PID %d = num %d running", os.getpid(), number)
            assert not potato.exists(), "We see someone else holding the potato file"

            # Put our name there
            with potato.open("w") as out_stream:
                out_stream.write(str(number))

            # Wait
            time.sleep(random.random() * 0.01)

            # Make sure our name is still there
            with potato.open() as in_stream:
                seen = in_stream.read().rstrip()
                assert seen == str(
                    number
                ), f"We are {number} but {seen} stole our potato!"

            potato.unlink()
            assert not potato.exists(), "We left the potato behind"
            log.info("PID %d = num %d dropped potato", os.getpid(), number)
        return True
    except:
        traceback.print_exc()
        return False


def _testLastProcessStandingTask(scope: Path, arena_name: str, number: int) -> bool:
    try:
        arena = LastProcessStandingArena(scope, arena_name)

        arena.enter()
        log.info("PID %d = num %d entered arena", os.getpid(), number)
        try:
            # We all make files
            my_precious = scope / f"precious{number}"

            # Put our name there
            with my_precious.open("w") as out_stream:
                out_stream.write(str(number))

            # Wait
            time.sleep(random.random() * 0.01)

            # Make sure our file is still there unmodified
            assert my_precious.exists(), f"Precious file {my_precious} has been stolen!"
            with my_precious.open() as in_stream:
                seen = in_stream.read().rstrip()
                assert seen == str(
                    number
                ), f"We are {number} but saw {seen} in our precious file!"
        finally:
            was_last = False
            for _ in arena.leave():
                log.info("PID %d = num %d is last standing", os.getpid(), number)

                # Clean up all the files
                for filename in os.listdir(scope):
                    if filename.startswith("precious"):
                        log.info(
                            "PID %d = num %d cleaning up %s",
                            os.getpid(),
                            number,
                            filename,
                        )
                        (scope / filename).unlink()

            log.info("PID %d = num %d left arena", os.getpid(), number)

        return True
    except:
        traceback.print_exc()
        return False



import math
import time
from multiprocessing import Process

from toil.common import Toil
from toil.job import Job


def think(seconds):
    start = time.time()
    while time.time() - start < seconds:
        # Use CPU
        math.sqrt(123456)


class TimeWaster(Job):
    def __init__(self, time_to_think, time_to_waste, space_to_waste, *args, **kwargs):
        self.time_to_think = time_to_think
        self.time_to_waste = time_to_waste
        self.space_to_waste = space_to_waste
        super().__init__(*args, **kwargs)

    def run(self, fileStore):
        # Waste some space
        file_path = fileStore.getLocalTempFile()
        with open(file_path, "w") as stream:
            for i in range(self.space_to_waste):
                stream.write("X")

        # Do some "useful" compute
        processes = []
        for core_number in range(max(1, self.cores)):
            # Use all the assigned cores to think
            p = Process(target=think, args=(self.time_to_think,))
            p.start()
            processes.append(p)
        for p in processes:
            p.join()

        # Also waste some time
        time.sleep(self.time_to_waste)


def main():
    options = Job.Runner.getDefaultArgumentParser().parse_args()

    job1 = TimeWaster(0, 0, 0, displayName="doNothing")
    job2 = TimeWaster(10, 0, 4096, displayName="efficientJob")
    job3 = TimeWaster(10, 0, 1024, cores=4, displayName="multithreadedJob")
    job4 = TimeWaster(1, 9, 65536, displayName="inefficientJob")

    job1.addChild(job2)
    job1.addChild(job3)
    job3.addChild(job4)

    with Toil(options) as toil:
        if not toil.options.restart:
            toil.start(job1)
        else:
            toil.restart()


if __name__ == "__main__":
    main()



from typing import TextIO

from _typeshed import Incomplete

from ..core.topictreetraverser import ITopicTreeVisitor, TopicTreeTraverser

class TopicTreePrinter(ITopicTreeVisitor):
    allowedExtras: Incomplete
    ALL_TOPICS_NAME: str
    def __init__(
        self,
        extra: Incomplete | None = ...,
        width: int = ...,
        indentStep: int = ...,
        bulletTopic: str = ...,
        bulletTopicItem: str = ...,
        bulletTopicArg: str = ...,
        fileObj: TextIO = ...,
    ) -> None: ...
    def getOutput(self): ...

def printTreeDocs(
    rootTopic: Incomplete | None = ..., topicMgr: Incomplete | None = ..., **kwargs
) -> None: ...



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import subprocess

from toil.statsAndLogging import logger, root_logger, set_logging_from_options
from toil.test import get_temp_file


# used by cactus
# TODO: only used in utilsTest.py; move this there once out of cactus
def system(command: list[str]) -> None:
    """
    A convenience wrapper around subprocess.check_call that logs the command before passing it
    on. The command can be either a string or a sequence of strings. If it is a string shell=True
    will be passed to subprocess.check_call.
    :type command: str | sequence[string]
    """
    logger.warning(
        'Deprecated toil method that will be moved/replaced in a future release."'
    )
    logger.debug(f"Running: {command}")
    subprocess.check_call(command, shell=isinstance(command, str), bufsize=-1)


# Used by cactus; now a wrapper and not used in Toil.
# TODO: Remove from cactus and then remove from Toil.
def getLogLevelString(logger=None):
    root_logger.warning(
        'Deprecated toil method.  Please call "logging.getLevelName" directly.'
    )
    if logger is None:
        logger = root_logger
    return logging.getLevelName(logger.getEffectiveLevel())


# Used by cactus; now a wrapper and not used in Toil.
# TODO: Remove from cactus and then remove from Toil.
def setLoggingFromOptions(options):
    logger.warning(
        'Deprecated toil method.  Please use "toil.statsAndLogging.set_logging_from_options()" instead."'
    )
    set_logging_from_options(options)


# Used by cactus; now a wrapper and not used in Toil.
# TODO: Remove from cactus and then remove from Toil.
def getTempFile(suffix="", rootDir=None):
    logger.warning(
        'Deprecated toil method.  Please use "toil.test.get_temp_file()" instead."'
    )
    return get_temp_file(suffix, rootDir)



from types import ModuleType
from typing import Any, Callable, Literal, Mapping, Sequence

from _typeshed import Incomplete

from .callables import CallArgsInfo as CallArgsInfo
from .callables import ListenerMismatchError as ListenerMismatchError
from .callables import UserListener as UserListener

class Topic: ...

class IListenerExcHandler:
    def __call__(self, listenerID: str, topicObj: Topic) -> None: ...

class Listener:
    AUTO_TOPIC: Incomplete
    acceptsAllKwargs: Incomplete
    curriedArgs: Incomplete
    def __init__(
        self,
        callable_obj: UserListener,
        argsInfo: CallArgsInfo,
        curriedArgs: Mapping[str, Any] = ...,
        onDead: Callable[[Listener], None] = ...,
    ) -> None: ...
    def name(self) -> str: ...
    def typeName(self) -> str: ...
    def module(self) -> ModuleType: ...
    def getCallable(self) -> UserListener: ...
    def isDead(self) -> bool: ...
    def wantsTopicObjOnCall(self) -> bool: ...
    def wantsAllMessageData(self) -> bool: ...
    def setCurriedArgs(self, **curriedArgs: Any) -> None: ...
    def __call__(
        self,
        kwargs: Mapping[str, Any],
        actualTopic: Topic,
        allKwargs: Mapping[str, Any] = ...,
    ) -> Literal[True]: ...

class ListenerValidator:
    def __init__(
        self, topicArgs: Sequence[str], topicKwargs: Sequence[str]
    ) -> None: ...
    def validate(
        self, listener: UserListener, curriedArgNames: Sequence[str] = ...
    ) -> CallArgsInfo: ...
    def isValid(
        self, listener: UserListener, curriedArgNames: Sequence[str] = ...
    ) -> bool: ...



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
import subprocess
from abc import abstractmethod
from uuid import uuid4

import pytest

from toil.test import (
    ToilTest,
    get_data,
    integrative,
    needs_fetchable_appliance,
    needs_google_project,
    needs_google_storage,
    slow,
    timeLimit,
)
from toil.version import exactPython

log = logging.getLogger(__name__)


@needs_google_project
@needs_google_storage
@integrative
@needs_fetchable_appliance
@slow
class AbstractGCEAutoscaleTest(ToilTest):
    projectID = os.getenv("TOIL_GOOGLE_PROJECTID")

    def sshUtil(self, command):
        baseCommand = ["toil", "ssh-cluster", "--insecure", "-p=gce", self.clusterName]
        callCommand = baseCommand + command
        subprocess.check_call(callCommand)

    def rsyncUtil(self, src, dest):
        baseCommand = [
            "toil",
            "rsync-cluster",
            "--insecure",
            "-p=gce",
            self.clusterName,
        ]
        callCommand = baseCommand + [src, dest]
        subprocess.check_call(callCommand)

    def destroyClusterUtil(self):
        callCommand = ["toil", "destroy-cluster", "-p=gce", self.clusterName]
        subprocess.check_call(callCommand)

    def createClusterUtil(self, args=None):
        if args is None:
            args = []
        callCommand = [
            "toil",
            "launch-cluster",
            self.clusterName,
            "-p=gce",
            "--keyPairName=%s" % self.keyName,
            "--leaderNodeType=%s" % self.leaderInstanceType,
            "--zone=%s" % self.googleZone,
        ]
        if self.botoDir is not None:
            callCommand += ["--boto=%s" % self.botoDir]
        callCommand = callCommand + args if args else callCommand
        log.info("createClusterUtil: %s" % "".join(callCommand))
        subprocess.check_call(callCommand)

    def cleanJobStoreUtil(self):
        callCommand = ["toil", "clean", self.jobStore]
        subprocess.check_call(callCommand)

    def __init__(self, methodName):
        super().__init__(methodName=methodName)
        # TODO: add TOIL_GOOGLE_KEYNAME to needs_google_project or ssh with SA account
        self.keyName = os.getenv("TOIL_GOOGLE_KEYNAME")
        # TODO: remove this when switching to google jobstore
        self.botoDir = os.getenv("TOIL_BOTO_DIR")
        # TODO: get this from SA account or add an environment variable
        self.googleZone = "us-west1-a"

        self.leaderInstanceType = "n1-standard-1"
        self.instanceTypes = ["n1-standard-2"]
        self.numWorkers = ["2"]
        self.numSamples = 2
        self.spotBid = 0.15

    def setUp(self):
        super().setUp()

    def tearDown(self):
        super().tearDown()
        self.destroyClusterUtil()
        self.cleanJobStoreUtil()

    # def getMatchingRoles(self, clusterName):
    #    ctx = AWSProvisioner._buildContext(clusterName)
    #    roles = list(ctx.local_roles())
    #    return roles

    def launchCluster(self):
        self.createClusterUtil()

    @abstractmethod
    def _getScript(self):
        """
        Download the test script needed by the inheriting unit test class.
        """
        raise NotImplementedError()

    @abstractmethod
    def _runScript(self, toilOptions):
        """
        Modify the provided Toil options to suit the test Toil script, then run the script with
        those arguments.

        :param toilOptions: List of Toil command line arguments. This list may need to be
               modified to suit the test script's requirements.
        """
        raise NotImplementedError()

    def _test(self, preemptibleJobs=False):
        """
        Does the work of the testing. Many features' test are thrown in here is no particular
        order
        """
        self.launchCluster()

        # TODO: What is the point of this test?
        # assert len(self.getMatchingRoles(self.clusterName)) == 1

        # TODO: Add a check of leader and node storage size if set.

        # --never-download prevents silent upgrades to pip, wheel and setuptools
        venv_command = [
            "virtualenv",
            "--system-site-packages",
            "--never-download",
            "--python",
            exactPython,
            "/home/venv",
        ]
        self.sshUtil(venv_command)

        self._getScript()

        toilOptions = [
            self.jobStore,
            "--batchSystem=mesos",
            "--workDir=/var/lib/toil",
            "--clean=always",
            "--retryCount=2",
            "--clusterStats=/home/",
            "--logDebug",
            "--logFile=/home/sort.log",
            "--provisioner=gce",
        ]

        toilOptions.extend(
            [
                "--nodeTypes=" + ",".join(self.instanceTypes),
                "--maxNodes=%s" % ",".join(self.numWorkers),
            ]
        )
        if preemptibleJobs:
            toilOptions.extend(["--defaultPreemptible"])

        self._runScript(toilOptions)

        # TODO: Does this just check if it is still running?
        # assert len(self.getMatchingRoles(self.clusterName)) == 1

        checkStatsCommand = [
            "/home/venv/bin/python",
            "-c",
            "import json; import os; "
            'json.load(open("/home/" + [f for f in os.listdir("/home/") '
            'if f.endswith(".json")].pop()))',
        ]

        self.sshUtil(checkStatsCommand)

        # TODO: Add a check to make sure everything is cleaned up.


@pytest.mark.timeout(1600)
class GCEAutoscaleTest(AbstractGCEAutoscaleTest):

    def __init__(self, name):
        super().__init__(name)
        self.clusterName = "provisioner-test-" + str(uuid4())
        self.requestedLeaderStorage = 80

    def setUp(self):
        super().setUp()
        self.jobStore = f"google:{self.projectID}:autoscale-{uuid4()}"

    def _getScript(self):
        # TODO: Isn't this the key file?
        fileToSort = os.path.join(os.getcwd(), str(uuid4()))
        with open(fileToSort, "w") as f:
            # Fixme: making this file larger causes the test to hang
            f.write("01234567890123456789012345678901")
        with get_data("test/sort/sort.py") as sort_py:
            self.rsyncUtil(
                sort_py,
                ":/home/sort.py",
            )
        self.rsyncUtil(fileToSort, ":/home/sortFile")
        os.unlink(fileToSort)

    def _runScript(self, toilOptions):
        runCommand = [
            "/home/venv/bin/python",
            "/home/sort.py",
            "--fileToSort=/home/sortFile",
        ]
        #'--sseKey=/home/sortFile']
        runCommand.extend(toilOptions)
        log.info("_runScript: %s" % "".join(runCommand))
        self.sshUtil(runCommand)

    def launchCluster(self):
        # add arguments to test that we can specify leader storage
        self.createClusterUtil(
            args=["--leaderStorage", str(self.requestedLeaderStorage)]
        )

    # TODO: aren't these checks inherited?
    @integrative
    @needs_google_project
    @needs_google_storage
    def testAutoScale(self):
        self.instanceTypes = ["n1-standard-2"]
        self.numWorkers = ["2"]
        self._test()

    @integrative
    @needs_google_project
    @needs_google_storage
    def testSpotAutoScale(self):
        self.instanceTypes = ["n1-standard-2:%f" % self.spotBid]
        # Some spot workers have a stopped state after being started, strangely.
        # This could be the natural preemption process, but it seems too rapid.
        self.numWorkers = ["3"]  # Try 3 to account for a stopped node.
        self._test(preemptibleJobs=True)


@pytest.mark.timeout(1600)
class GCEStaticAutoscaleTest(GCEAutoscaleTest):
    """
    Runs the tests on a statically provisioned cluster with autoscaling enabled.
    """

    def __init__(self, name):
        super().__init__(name)
        self.requestedNodeStorage = 20

    def launchCluster(self):
        self.createClusterUtil(
            args=[
                "--leaderStorage",
                str(self.requestedLeaderStorage),
                "--nodeTypes",
                ",".join(self.instanceTypes),
                "-w",
                ",".join(self.numWorkers),
                "--nodeStorage",
                str(self.requestedLeaderStorage),
            ]
        )

        # TODO: check the number of workers and their storage
        # nodes = AWSProvisioner._getNodesInCluster(ctx, self.clusterName, both=True)
        # nodes.sort(key=lambda x: x.launch_time)
        # assuming that leader is first
        # workers = nodes[1:]
        # test that two worker nodes were created
        # self.assertEqual(2, len(workers))
        # test that workers have expected storage size
        # just use the first worker
        # worker = workers[0]
        # worker = next(wait_instances_running(ctx.ec2, [worker]))
        # rootBlockDevice = worker.block_device_mapping["/dev/xvda"]
        # self.assertTrue(isinstance(rootBlockDevice, BlockDeviceType))
        # rootVolume = ctx.ec2.get_all_volumes(volume_ids=[rootBlockDevice.volume_id])[0]
        # self.assertGreaterEqual(rootVolume.size, self.requestedNodeStorage)

    def _runScript(self, toilOptions):
        runCommand = [
            "/home/venv/bin/python",
            "/home/sort.py",
            "--fileToSort=/home/sortFile",
        ]
        runCommand.extend(toilOptions)
        log.info("_runScript: %s" % "".join(runCommand))
        self.sshUtil(runCommand)


@pytest.mark.timeout(1800)
class GCEAutoscaleTestMultipleNodeTypes(AbstractGCEAutoscaleTest):

    def __init__(self, name):
        super().__init__(name)
        self.clusterName = "provisioner-test-" + str(uuid4())

    def setUp(self):
        super().setUp()
        self.jobStore = f"google:{self.projectID}:multinode-{uuid4()}"

    def _getScript(self):
        sseKeyFile = os.path.join(os.getcwd(), "keyFile")
        with open(sseKeyFile, "w") as f:
            f.write("01234567890123456789012345678901")
        with get_data("test/sort/sort.py") as sort_py:
            self.rsyncUtil(
                sort_py,
                ":/home/sort.py",
            )
        self.rsyncUtil(sseKeyFile, ":/home/keyFile")
        os.unlink(sseKeyFile)

    def _runScript(self, toilOptions):
        # Set memory requirements so that sort jobs can be run
        # on small instances, but merge jobs must be run on large
        # instances
        runCommand = [
            "/home/venv/bin/python",
            "/home/sort.py",
            "--fileToSort=/home/s3am/bin/asadmin",
            "--sortMemory=0.6G",
            "--mergeMemory=3.0G",
        ]
        runCommand.extend(toilOptions)
        # runCommand.append('--sseKey=/home/keyFile')
        log.info("_runScript: %s" % "".join(runCommand))
        self.sshUtil(runCommand)

    @integrative
    @needs_google_project
    @needs_google_storage
    def testAutoScale(self):
        self.instanceTypes = ["n1-standard-2", "n1-standard-4"]
        self.numWorkers = ["2", "1"]
        self._test()


@pytest.mark.timeout(1800)
class GCERestartTest(AbstractGCEAutoscaleTest):
    """
    This test insures autoscaling works on a restarted Toil run
    """

    def __init__(self, name):
        super().__init__(name)
        self.clusterName = "restart-test-" + str(uuid4())

    def setUp(self):
        super().setUp()
        self.instanceTypes = ["n1-standard-1"]
        self.numWorkers = ["1"]
        self.scriptName = "/home/restartScript.py"
        # TODO: replace this with a google job store
        zone = "us-west-2"
        self.jobStore = f"google:{self.projectID}:restart-{uuid4()}"

    def _getScript(self):
        with get_data("test/provisioners/restartScript.py") as restartScript:
            self.rsyncUtil(
                restartScript,
                ":" + self.scriptName,
            )

    def _runScript(self, toilOptions):
        # clean = onSuccess
        disallowedOptions = ["--clean=always", "--retryCount=2"]
        newOptions = [
            option for option in toilOptions if option not in disallowedOptions
        ]
        try:
            # include a default memory - on restart the minimum memory requirement is the default, usually 2 GB
            command = [
                "/home/venv/bin/python",
                self.scriptName,
                "-e",
                "FAIL=true",
                "--defaultMemory=50000000",
            ]
            command.extend(newOptions)
            self.sshUtil(command)
        except subprocess.CalledProcessError:
            pass
        else:
            self.fail("Command succeeded when we expected failure")
        with timeLimit(1200):
            command = [
                "/home/venv/bin/python",
                self.scriptName,
                "--restart",
                "--defaultMemory=50000000",
            ]
            command.extend(toilOptions)
            self.sshUtil(command)

    @integrative
    def testAutoScaledCluster(self):
        self._test()



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import time
from typing import Optional

from toil.bus import JobUpdatedMessage, MessageBus
from toil.job import CheckpointJobDescription, JobDescription
from toil.jobStores.abstractJobStore import AbstractJobStore, NoSuchJobException

logger = logging.getLogger(__name__)


class ToilState:
    """
    Holds the leader's scheduling information.

    But onlt that which does not need to be persisted back to the JobStore (such
    as information on completed and outstanding predecessors)

    Holds the true single copies of all JobDescription objects that the Leader
    and ServiceManager will use. The leader and service manager shouldn't do
    their own load() and update() calls on the JobStore; they should go through
    this class.

    Everything in the leader should reference JobDescriptions by ID.

    Only holds JobDescription objects, not Job objects, and those
    JobDescription objects only exist in single copies.
    """

    def __init__(
        self,
        jobStore: AbstractJobStore,
    ) -> None:
        """
        Create an empty ToilState over the given job store.

        After calling this, you probably want to call load_workflow() to
        initialize the state given the root job, correct jobs to reflect a
        consistent state, and find jobs that need leader attention.

        :param jobStore: The job store to use.
        """
        # We have a public bus for messages about job state changes.
        self.bus = MessageBus()

        # We need to keep the job store so we can load and save jobs.
        self.__job_store = jobStore

        # This holds the one true copy of every JobDescription in the leader.
        # TODO: Do in-place update instead of assignment when we load so we
        # can't let any non-true copies escape.
        self.__job_database: dict[str, JobDescription] = {}

        # Maps from successor (child or follow-on) jobStoreID to predecessor jobStoreIDs
        self.successor_to_predecessors: dict[str, set[str]] = {}

        # Hash of jobStoreIDs to counts of numbers of successors issued.
        # There are no entries for jobs without successors in this map.
        self.successorCounts: dict[str, int] = {}

        # This is a hash of service jobs, referenced by jobStoreID, to their client's ID
        self.service_to_client: dict[str, str] = {}

        # Holds, for each client job ID, the job IDs of its services that are
        # possibly currently issued. Includes every service host that has been
        # given to the service manager by the leader, and hasn't been seen by
        # the leader as stopped yet.
        self.servicesIssued: dict[str, set[str]] = {}

        # Holds the IDs of jobs that are currently issued to the batch system
        # and haven't come back yet.
        # TODO: a bit redundant with leader's issued_jobs_by_batch_system_id
        self.jobs_issued: set[str] = set()

        # The set of totally failed jobs - this needs to be filtered at the
        # end to remove jobs that were removed by checkpoints
        self.totalFailedJobs: set[str] = set()

        # Jobs (as jobStoreIDs) with successors that have totally failed
        self.hasFailedSuccessors: set[str] = set()

        # The set of successors of failed jobs as a set of jobStoreIds
        self.failedSuccessors: set[str] = set()

        # Set of jobs that have multiple predecessors that have one or more predecessors
        # finished, but not all of them.
        self.jobsToBeScheduledWithMultiplePredecessors: set[str] = set()

    def load_workflow(
        self,
        rootJob: JobDescription,
        jobCache: Optional[dict[str, JobDescription]] = None,
    ) -> None:
        """
        Load the workflow rooted at the given job.

        If jobs are loaded that have updated and need to be dealt with by the
        leader, JobUpdatedMessage messages will be sent to the message bus.

        The jobCache is a map from jobStoreID to JobDescription or None. Is
        used to speed up the building of the state when loading initially from
        the JobStore, and is not preserved.

        :param rootJob: The description for the root job of the workflow being run.

        :param jobCache: A dict to cache downloaded job descriptions in, keyed by ID.
        """
        if jobCache is not None:
            # Load any pre-cached JobDescriptions we were given
            self.__job_database.update(jobCache)

        # Build the state from the jobs
        self._buildToilState(rootJob)

    def job_exists(self, job_id: str) -> bool:
        """
        Test if the givin job exists now.

        Returns True if the given job exists right now, and false if it hasn't
        been created or it has been deleted elsewhere.

        Doesn't guarantee that the job will or will not be gettable, if racing
        another process, or if it is still cached.
        """
        return self.__job_store.job_exists(job_id)

    def get_job(self, job_id: str) -> JobDescription:
        """Get the one true copy of the JobDescription with the given ID."""
        if job_id not in self.__job_database:
            # Go get the job for the first time
            self.__job_database[job_id] = self.__job_store.load_job(job_id)
        return self.__job_database[job_id]

    def commit_job(self, job_id: str) -> None:
        """
        Save back any modifications made to a JobDescription.

        (one retrieved from get_job())
        """
        self.__job_store.update_job(self.__job_database[job_id])

    def delete_job(self, job_id: str) -> None:
        """
        Destroy a JobDescription.

        May raise an exception if the job could not be cleaned up (i.e. files
        belonging to it failed to delete).
        """
        # Do the backing delete first
        self.__job_store.delete_job(job_id)
        # If that succeeds, drop from cache
        if job_id in self.__job_database:
            del self.__job_database[job_id]

    def reset_job(self, job_id: str) -> None:
        """
        Discard any local modifications to a JobDescription.

        Will make modifications from other hosts visible.
        """
        try:
            new_truth = self.__job_store.load_job(job_id)
        except NoSuchJobException:
            # The job is gone now.
            if job_id in self.__job_database:
                # So forget about it
                del self.__job_database[job_id]
                # TODO: Other collections may still reference it.
            return
        if job_id in self.__job_database:
            # Update the one true copy in place
            old_truth = self.__job_database[job_id]
            old_truth.assert_is_not_newer_than(new_truth)
            old_truth.__dict__.update(new_truth.__dict__)
        else:
            # Just keep the new one
            self.__job_database[job_id] = new_truth

    def reset_job_expecting_change(self, job_id: str, timeout: float) -> bool:
        """
        Discard any local modifications to a JobDescription.

        Will make modifications from other hosts visible.

        Will wait for up to timeout seconds for a modification (or deletion)
        from another host to actually be visible.

        Always replaces the JobDescription with what is stored in the job
        store, even if no modification ends up being visible.

        Returns True if an update was detected in time, and False otherwise.
        """

        start_time = time.time()
        wait_time = 0.1
        initially_known = job_id in self.__job_database
        new_truth: Optional[JobDescription] = None
        while True:
            try:
                new_truth = self.__job_store.load_job(job_id)
            except NoSuchJobException:
                # The job is gone now.
                if job_id in self.__job_database:
                    # So forget about it
                    del self.__job_database[job_id]
                    # TODO: Other collections may still reference it.
                if initially_known:
                    # Job was deleted, that's an update
                    return True
            else:
                if job_id in self.__job_database:
                    # We have an old version to compare against
                    old_truth = self.__job_database[job_id]
                    old_truth.assert_is_not_newer_than(new_truth)
                    if old_truth.is_updated_by(new_truth):
                        # Do the update
                        old_truth.__dict__.update(new_truth.__dict__)
                        return True
                else:
                    # Just keep the new one. That's an update.
                    self.__job_database[job_id] = new_truth
                    return True
            # We looked but didn't get a good update
            time_elapsed = time.time() - start_time
            if time_elapsed >= timeout:
                # We're out of time to check.
                if new_truth is not None:
                    # Commit whatever we managed to load to accomplish a real
                    # reset.
                    old_truth.__dict__.update(new_truth.__dict__)
                return False
            # Wait a little and poll again
            time.sleep(min(timeout - time_elapsed, wait_time))
            # Using exponential backoff
            wait_time *= 2

    # The next 3 functions provide tracking of how many successor jobs a given job
    # is waiting on, exposing only legit operations.
    # TODO: turn these into messages?
    def successors_pending(self, predecessor_id: str, count: int) -> None:
        """
        Remember that the given job has the given number more pending successors.

        (that have not yet succeeded or failed.)
        """
        if predecessor_id not in self.successorCounts:
            self.successorCounts[predecessor_id] = count

        else:
            self.successorCounts[predecessor_id] += count
        logger.debug(
            "Successors: %d more for %s, now have %d",
            count,
            predecessor_id,
            self.successorCounts[predecessor_id],
        )

    def successor_returned(self, predecessor_id: str) -> None:
        """
        Remember that the given job has one fewer pending successors.

        (because one has succeeded or failed.)
        """
        if predecessor_id not in self.successorCounts:
            raise RuntimeError(
                f"Tried to remove successor of {predecessor_id} that wasn't added!"
            )
        else:
            self.successorCounts[predecessor_id] -= 1
            logger.debug(
                "Successors: one fewer for %s, now have %d",
                predecessor_id,
                self.successorCounts[predecessor_id],
            )
            if self.successorCounts[predecessor_id] == 0:
                del self.successorCounts[predecessor_id]

    def count_pending_successors(self, predecessor_id: str) -> int:
        """
        Count number of pending successors of the given job.

        Pending successors are those which have not yet succeeded or failed.
        """
        if predecessor_id not in self.successorCounts:
            return 0
        else:
            return self.successorCounts[predecessor_id]

    def _buildToilState(self, jobDesc: JobDescription) -> None:
        """
        Build the ToilState class from the subtree root JobDescription.

        Updated jobs that the leader needs to deal with will have messages sent
        to the message bus.

        :param jobDesc: The description for the root job of the workflow being run.
        """
        # If the job description has a body, is a checkpoint, has services
        # or is ready to be deleted it is ready to be processed (i.e. it is updated)
        if (
            jobDesc.has_body()
            or (
                isinstance(jobDesc, CheckpointJobDescription)
                and jobDesc.checkpoint is not None
            )
            or len(jobDesc.services) > 0
            or jobDesc.nextSuccessors() is None
        ):
            logger.debug(
                "Found job to run: %s, with body: %s, with checkpoint: %s, with "
                "services: %s, with no next successors: %s",
                jobDesc.jobStoreID,
                jobDesc.has_body(),
                isinstance(jobDesc, CheckpointJobDescription)
                and jobDesc.checkpoint is not None,
                len(jobDesc.services) > 0,
                jobDesc.nextSuccessors() is None,
            )
            # Set the job updated because we should be able to make progress on it.
            self.bus.publish(JobUpdatedMessage(str(jobDesc.jobStoreID), 0))

            if (
                isinstance(jobDesc, CheckpointJobDescription)
                and jobDesc.checkpoint is not None
            ):
                jobDesc.restore_checkpoint()

        else:  # There exist successors
            logger.debug(
                "Adding job: %s to the state with %s successors",
                jobDesc.jobStoreID,
                len(jobDesc.nextSuccessors() or set()),
            )

            # Record the number of successors
            self.successorCounts[str(jobDesc.jobStoreID)] = len(
                jobDesc.nextSuccessors() or set()
            )

            def processSuccessorWithMultiplePredecessors(
                successor: JobDescription,
            ) -> None:
                # If jobDesc is not reported as complete by the successor
                if jobDesc.jobStoreID not in successor.predecessorsFinished:

                    # Update the successor's status to mark the predecessor complete
                    successor.predecessorsFinished.add(jobDesc.jobStoreID)

                # If the successor has no predecessors to finish
                if len(successor.predecessorsFinished) > successor.predecessorNumber:
                    raise RuntimeError(
                        "There are more finished predecessors than possible."
                    )
                if len(successor.predecessorsFinished) == successor.predecessorNumber:

                    # It is ready to be run, so remove it from the set of waiting jobs
                    self.jobsToBeScheduledWithMultiplePredecessors.remove(
                        successorJobStoreID
                    )

                    # Recursively consider the successor
                    self._buildToilState(successor)

            # For each successor
            for successorJobStoreID in jobDesc.nextSuccessors() or set():

                # If the successor does not yet point back at a
                # predecessor we have not yet considered it
                if successorJobStoreID not in self.successor_to_predecessors:

                    # Add the job as a predecessor
                    self.successor_to_predecessors[successorJobStoreID] = {
                        str(jobDesc.jobStoreID)
                    }

                    # We load the successor job
                    successor = self.get_job(successorJobStoreID)

                    # If predecessor number > 1 then the successor has multiple predecessors
                    if successor.predecessorNumber > 1:

                        # We put the successor job in the set of waiting successor
                        # jobs with multiple predecessors
                        if (
                            successorJobStoreID
                            in self.jobsToBeScheduledWithMultiplePredecessors
                        ):
                            raise RuntimeError(
                                "Failed to schedule the successor job. The successor job is already scheduled."
                            )
                        self.jobsToBeScheduledWithMultiplePredecessors.add(
                            successorJobStoreID
                        )

                        # Process successor
                        processSuccessorWithMultiplePredecessors(successor)

                    else:
                        # The successor has only this job as a predecessor so
                        # recursively consider the successor
                        self._buildToilState(successor)

                else:
                    # We've already seen the successor

                    # Add the job as a predecessor
                    if (
                        jobDesc.jobStoreID
                        in self.successor_to_predecessors[successorJobStoreID]
                    ):
                        raise RuntimeError(
                            "Failed to add the job as a predecessor. The job is already added as a predecessor."
                        )
                    self.successor_to_predecessors[successorJobStoreID].add(
                        str(jobDesc.jobStoreID)
                    )

                    # If the successor has multiple predecessors
                    if (
                        successorJobStoreID
                        in self.jobsToBeScheduledWithMultiplePredecessors
                    ):

                        # Get the successor from cache
                        successor = self.get_job(successorJobStoreID)

                        # Process successor
                        processSuccessorWithMultiplePredecessors(successor)



from argparse import ArgumentParser

from configargparse import SUPPRESS

from toil.lib.conversions import strtobool


def add_wdl_options(parser: ArgumentParser, suppress: bool = True) -> None:
    """
    Add WDL options to a parser. This only adds nonpositional WDL arguments
    :param parser: Parser to add options to
    :param suppress: Suppress help output
    :return: None
    """
    suppress_help = SUPPRESS if suppress else None
    # include arg names without a wdl specifier if suppress is False
    # this is to avoid possible duplicate options in custom toil scripts, ex outputFile can be a common argument name
    # TODO: Why do we even need them at all in other Toil scripts? Do we have to worry about dest= collisions?
    # TODO: Can the better option name be first?
    output_dialect_arguments = ["--wdlOutputDialect"] + (
        ["--outputDialect"] if not suppress else []
    )
    parser.add_argument(
        *output_dialect_arguments,
        dest="output_dialect",
        type=str,
        default="cromwell",
        choices=["cromwell", "miniwdl"],
        help=suppress_help
        or (
            "JSON output format dialect. 'cromwell' just returns the workflow's "
            "output values as JSON, while 'miniwdl' nests that under an 'outputs' "
            "key, and includes a 'dir' key where files are written."
        )
    )
    output_directory_arguments = ["--wdlOutputDirectory"] + (
        ["--outputDirectory", "-o"] if not suppress else []
    )
    parser.add_argument(
        *output_directory_arguments,
        dest="output_directory",
        type=str,
        default=None,
        help=suppress_help
        or (
            "Directory or URI prefix to save output files at. By default a new directory is created "
            "in the current directory."
        )
    )
    output_file_arguments = ["--wdlOutputFile"] + (
        ["--outputFile", "-m"] if not suppress else []
    )
    parser.add_argument(
        *output_file_arguments,
        dest="output_file",
        type=str,
        default=None,
        help=suppress_help or "File or URI to save output JSON to."
    )
    reference_inputs_arguments = ["--wdlReferenceInputs"] + (
        ["--referenceInputs"] if not suppress else []
    )
    parser.add_argument(
        *reference_inputs_arguments,
        dest="reference_inputs",
        type=strtobool,
        default=False,
        help=suppress_help or "Pass input files by URL"
    )
    container_arguments = ["--wdlContainer"] + (["--container"] if not suppress else [])
    parser.add_argument(
        *container_arguments,
        dest="container",
        type=str,
        choices=["singularity", "docker", "auto"],
        default="auto",
        help=suppress_help or "Container engine to use to run WDL tasks"
    )
    all_call_outputs_arguments = ["--wdlAllCallOutputs"] + (
        ["--allCallOutputs"] if not suppress else []
    )
    parser.add_argument(
        *all_call_outputs_arguments,
        dest="all_call_outputs",
        type=strtobool,
        default=None,
        help=suppress_help or "Keep and return all call outputs as workflow outputs"
    )

    strict_arguments = ["--wdlStrict"] + (
        ["--strict"] if not suppress else []
    )
    parser.add_argument(
        *strict_arguments,
        dest="strict",
        type=strtobool,
        default=False,
        help=suppress_help or "Exit runner if workflow has any lint warnings"
    )


# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Launches a toil leader instance with the specified provisioner."""

import logging
import os
from typing import Union

from toil import applianceSelf
from toil.common import parser_with_common_options

try:
    from toil.lib.aws import build_tag_dict_from_env
except ModuleNotFoundError:
    build_tag_dict_from_env: dict[str, str] = lambda _: {}  # type: ignore[no-redef]
from toil.lib.conversions import opt_strtobool
from toil.provisioners import check_valid_node_types, cluster_factory, parse_node_types
from toil.statsAndLogging import set_logging_from_options

logger = logging.getLogger(__name__)


def create_tags_dict(tags: list[str]) -> dict[str, str]:
    tags_dict = dict()
    for tag in tags:
        try:
            key, value = tag.split("=")
        except ValueError:
            logger.error("Tag specification '%s' must contain '='", tag)
            raise
        tags_dict[key] = value
    return tags_dict


def main() -> None:
    parser = parser_with_common_options(
        provisioner_options=True, jobstore_option=False, prog="toil launch-cluster"
    )
    parser.add_argument(
        "-T",
        "--clusterType",
        dest="clusterType",
        choices=["mesos", "kubernetes"],
        default=None,  # TODO: change default to "kubernetes" when we are ready.
        help="Cluster scheduler to use.",
    )
    parser.add_argument(
        "--leaderNodeType",
        dest="leaderNodeType",
        required=True,
        help="Non-preemptible node type to use for the cluster leader.",
    )
    parser.add_argument(
        "--keyPairName",
        dest="keyPairName",
        help="On AWS, the name of the AWS key pair to include on the instance."
        " On Google/GCE, this is the ssh key pair.",
    )
    parser.add_argument(
        "--owner",
        dest="owner",
        help="The owner tag for all instances. If not given, the value in"
        "TOIL_OWNER_TAG will be used, or else the value of --keyPairName.",
    )
    parser.add_argument(
        "--boto",
        dest="botoPath",
        help="The path to the boto credentials directory. This is transferred "
        "to all nodes in order to access the AWS jobStore from non-AWS instances.",
    )
    parser.add_argument(
        "-t",
        "--tag",
        metavar="NAME=VALUE",
        dest="tags",
        default=[],
        action="append",
        help="Tags are added to the AWS cluster for this node and all of its "
        "children. Tags are of the form:\n"
        " -t key1=value1 --tag key2=value2\n"
        "Multiple tags are allowed and each tag needs its own flag. By "
        "default the cluster is tagged with "
        " {\n"
        '      "Name": clusterName,\n'
        '      "Owner": IAM username\n'
        " }. ",
    )

    parser.add_argument(
        "--network", help="GCE cloud network to use. default: 'default'"
    )
    parser.add_argument(
        "--vpcSubnet",
        help="VPC subnet ID to launch cluster leader in. Uses default subnet "
        "if not specified. This subnet needs to have auto assign IPs turned on.",
    )
    parser.add_argument(
        "--use_private_ip",
        dest="use_private_ip",
        action="store_true",
        default=False,
        help="if specified, ignore the public ip of the nodes",
    )
    parser.add_argument(
        "--nodeTypes",
        dest="nodeTypes",
        default=None,
        type=str,
        help="Specifies a list of comma-separated node types, each of which is "
        "composed of slash-separated instance types, and an optional spot "
        "bid set off by a colon, making the node type preemptible. Instance "
        "types may appear in multiple node types, and the same node type "
        "may appear as both preemptible and non-preemptible.\n"
        "Valid argument specifying two node types:\n"
        "\tc5.4xlarge/c5a.4xlarge:0.42,t2.large\n"
        "Node types:\n"
        "\tc5.4xlarge/c5a.4xlarge:0.42 and t2.large\n"
        "Instance types:\n"
        "\tc5.4xlarge, c5a.4xlarge, and t2.large\n"
        "Semantics:\n"
        "\tBid $0.42/hour for either c5.4xlarge or c5a.4xlarge instances,\n"
        "\ttreated interchangeably, while they are available at that price,\n"
        "\tand buy t2.large instances at full price\n"
        "Must also provide the --workers argument to specify how many "
        "workers of each node type to create.",
    )
    parser.add_argument(
        "-w",
        "--workers",
        dest="workers",
        default=None,
        type=str,
        help="Comma-separated list of the ranges of numbers of workers of each "
        "node type to launch, such as '0-2,5,1-3'. If a range is given, "
        "workers will automatically be launched and terminated by the cluster "
        "to auto-scale to the workload.",
    )
    parser.add_argument(
        "--leaderStorage",
        dest="leaderStorage",
        type=int,
        default=50,
        help="Specify the size (in gigabytes) of the root volume for the leader "
        "instance. This is an EBS volume.",
    )
    parser.add_argument(
        "--nodeStorage",
        dest="nodeStorage",
        type=int,
        default=50,
        help="Specify the size (in gigabytes) of the root volume for any worker "
        "instances created when using the -w flag. This is an EBS volume.",
    )
    parser.add_argument(
        "--forceDockerAppliance",
        dest="forceDockerAppliance",
        action="store_true",
        default=False,
        help="Disables sanity checking the existence of the docker image specified "
        "by TOIL_APPLIANCE_SELF, which Toil uses to provision mesos for "
        "autoscaling.",
    )
    parser.add_argument(
        "--awsEc2ProfileArn",
        dest="awsEc2ProfileArn",
        default=None,
        type=str,
        help="If provided, the specified ARN is used as the instance profile for EC2 instances."
        "Useful for setting custom IAM profiles. If not specified, a new IAM role is created "
        "by default with sufficient access to perform basic cluster operations.",
    )
    parser.add_argument(
        "--awsEc2ExtraSecurityGroupId",
        dest="awsEc2ExtraSecurityGroupIds",
        default=[],
        action="append",
        help="Any additional security groups to attach to EC2 instances. Note that a security group "
        "with its name equal to the cluster name will always be created, thus ensure that "
        "the extra security groups do not have the same name as the cluster name.",
    )
    parser.add_argument(
        "--allowFuse",
        type=opt_strtobool,
        default=True,
        help="Enable both the leader and worker nodes to be able to run Singularity with FUSE. For "
        "Kubernetes, this will make the leader privileged and ask workers to run as privileged. "
        "(default: %(default)s)",
    )
    # TODO Set Aws Profile in CLI options
    options = parser.parse_args()
    set_logging_from_options(options)

    tags = create_tags_dict(options.tags) if options.tags else build_tag_dict_from_env()

    # Get worker node types
    worker_node_types = parse_node_types(options.nodeTypes)
    check_valid_node_types(
        options.provisioner, worker_node_types + [({options.leaderNodeType}, None)]
    )

    # Holds string ranges, like "5", or "3-10"
    worker_node_ranges = options.workers.split(",") if options.workers else []

    # checks the validity of TOIL_APPLIANCE_SELF before proceeding
    applianceSelf(forceDockerAppliance=options.forceDockerAppliance)

    # This holds either ints to launch static nodes, or tuples of ints
    # specifying ranges to launch managed auto-scaling nodes, for each type.
    nodeCounts: list[Union[int, tuple[int, int]]] = []

    if (worker_node_types != [] or worker_node_ranges != []) and not (
        worker_node_types != [] and worker_node_ranges != []
    ):
        raise RuntimeError("The --nodeTypes option requires --workers, and visa versa.")
    if worker_node_types and worker_node_ranges:
        if not len(worker_node_types) == len(worker_node_ranges):
            raise RuntimeError(
                "List of worker count ranges must be the same length as the list of node types."
            )

        for spec in worker_node_ranges:
            if "-" in spec:
                # Provision via autoscaling
                parts = spec.split("-")
                if len(parts) != 2:
                    raise RuntimeError("Unacceptable range: " + spec)
                nodeCounts.append((int(parts[0]), int(parts[1])))
            else:
                # Provision fixed nodes
                nodeCounts.append(int(spec))

    owner = (
        options.owner or os.getenv("TOIL_OWNER_TAG") or options.keyPairName or "toil"
    )

    # Check to see if the user specified a zone. If not, see if one is stored in an environment variable.
    options.zone = options.zone or os.environ.get(
        f"TOIL_{options.provisioner.upper()}_ZONE"
    )

    if not options.zone:
        raise RuntimeError(
            f"Please provide a value for --zone or set a default in the "
            f"TOIL_{options.provisioner.upper()}_ZONE environment variable."
        )

    if options.clusterType == "mesos":
        logger.warning(
            "You are using a Mesos cluster, which is no longer recommended as Toil is "
            "transitioning to Kubernetes-based clusters. Consider switching to "
            "--clusterType=kubernetes instead."
        )

    if options.clusterType is None:
        logger.warning(
            'Argument --clusterType is not set... using "mesos". '
            "In future versions of Toil, the default cluster scheduler will be "
            'set to "kubernetes" if the cluster type is not specified.'
        )
        options.clusterType = "mesos"

    logger.info("Creating cluster %s...", options.clusterName)

    cluster = cluster_factory(
        provisioner=options.provisioner,
        clusterName=options.clusterName,
        clusterType=options.clusterType,
        zone=options.zone,
        nodeStorage=options.nodeStorage,
        enable_fuse=options.allowFuse,
    )

    cluster.launchCluster(
        leaderNodeType=options.leaderNodeType,
        leaderStorage=options.leaderStorage,
        owner=owner,
        keyName=options.keyPairName,
        botoPath=options.botoPath,
        userTags=tags,
        network=options.network,
        vpcSubnet=options.vpcSubnet,
        awsEc2ProfileArn=options.awsEc2ProfileArn,
        awsEc2ExtraSecurityGroupIds=options.awsEc2ExtraSecurityGroupIds,
    )

    for typeNum, spec in enumerate(nodeCounts):
        # For each batch of workers to make
        wanted = worker_node_types[typeNum]

        if isinstance(spec, int):
            # Make static nodes

            if spec == 0:
                # Don't make anything
                continue

            if wanted[1] is None:
                # Make non-spot instances
                cluster.addNodes(nodeTypes=wanted[0], numNodes=spec, preemptible=False)
            else:
                # We have a spot bid
                cluster.addNodes(
                    nodeTypes=wanted[0],
                    numNodes=spec,
                    preemptible=True,
                    spotBid=wanted[1],
                )

        elif isinstance(spec, tuple):
            # Make a range of auto-scaling nodes

            max_count, min_count = spec

            if max_count < min_count:
                # Flip them around
                min_count, max_count = max_count, min_count

            if max_count == 0:
                # Don't want any
                continue

            if wanted[1] is None:
                # Make non-spot instances
                cluster.addManagedNodes(
                    nodeTypes=wanted[0],
                    minNodes=min_count,
                    maxNodes=max_count,
                    preemptible=False,
                )
            else:
                # Bid at the given price.
                cluster.addManagedNodes(
                    nodeTypes=wanted[0],
                    minNodes=min_count,
                    maxNodes=max_count,
                    preemptible=True,
                    spotBid=wanted[1],
                )

    logger.info("Cluster created successfully.")



from textwrap import TextWrapper
from typing import Sequence, Tuple

from .topicexc import TopicNameError

UNDERSCORE: str
ALL_TOPICS: str

class WeakNone:
    def __call__(self) -> None: ...

def smartDedent(paragraph: str) -> str: ...
def validateName(topicName: str) -> None: ...
def stringize(topicName: Sequence[str]) -> str: ...
def tupleize(topicName: str) -> Tuple[str, ...]: ...



from typing import Any, Dict, Iterable, List, Mapping
from typing import Sequence as Seq
from typing import Tuple

from _typeshed import Incomplete

from .annotations import annotationType
from .listener import UserListener
from .topicexc import MessageDataSpecError as MessageDataSpecError
from .topicutils import WeakNone, stringize

ArgsDocs = Dict[str, str]
MsgData = Mapping[str, Any]

def verifyArgsDifferent(
    allArgs: Iterable[Any], allParentArgs: Iterable[Any], topicName: str
) -> None: ...
def verifySubset(
    all: Iterable[Any], sub: Iterable[Any], topicName: str, extraMsg: str = ...
) -> None: ...
def topicArgsFromCallable(
    _callable: UserListener, ignoreArgs: Seq[str] = ...
) -> Tuple[ArgsDocs, List[str]]: ...

class ArgSpecGiven:
    SPEC_GIVEN_NONE: int
    SPEC_GIVEN_ALL: int
    reqdArgs: Incomplete
    argsSpecType: Incomplete
    argsDocs: Incomplete
    def __init__(self, argsDocs: ArgsDocs = ..., reqdArgs: Seq[str] = ...) -> None: ...
    def setAll(self, allArgsDocs: ArgsDocs, reqdArgs: Seq[str] = ...) -> None: ...
    def isComplete(self) -> bool: ...
    def getOptional(self) -> List[str]: ...

class SenderMissingReqdMsgDataError(RuntimeError):
    def __init__(
        self, topicName: str, argNames: Seq[str], missing: Seq[str]
    ) -> None: ...

class SenderUnknownMsgDataError(RuntimeError):
    def __init__(self, topicName: str, argNames: Seq[str], extra: Seq[str]) -> None: ...

class ArgsInfo:
    SPEC_MISSING: int
    SPEC_COMPLETE: int
    topicNameTuple: Incomplete
    allOptional: Incomplete
    allDocs: Incomplete
    allRequired: Incomplete
    argsSpecType: Incomplete
    parentAI: Incomplete
    childrenAI: Incomplete
    argsAddedToParent: Incomplete
    def __init__(
        self,
        topicNameTuple: Seq[str],
        specGiven: ArgSpecGiven,
        parentArgsInfo: ArgsInfo,
    ) -> None: ...
    def isComplete(self) -> bool: ...
    def getArgs(self) -> List[str]: ...
    def numArgs(self) -> int: ...
    def getReqdArgs(self) -> List[str]: ...
    def getOptArgs(self) -> List[str]: ...
    def getArgsDocs(self) -> ArgsDocs: ...
    def setArgsDocs(self, docs: ArgsDocs) -> None: ...
    def check(self, msgData: MsgData) -> None: ...
    def filterArgs(self, msgData: MsgData) -> MsgData: ...
    def hasSameArgs(self, *argNames: Seq[str]) -> bool: ...
    def hasParent(self, argsInfo: ArgsInfo) -> bool: ...
    def getCompleteAI(self) -> ArgsInfo: ...
    def updateAllArgsFinal(self, topicDefn: ArgSpecGiven) -> None: ...



# Copyright (C) 2024 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains functions for integrating Toil with GA4GH Tool Registry Service
servers, for fetching workflows.
"""

import hashlib
import logging
import os
import shutil
import sys
import tempfile
import zipfile
from typing import Any, Literal, Optional, Union, TypedDict, cast

from urllib.parse import urlparse, unquote, quote
import requests

from toil.lib.retry import retry
from toil.lib.io import file_digest, robust_rmtree
from toil.lib.web import web_session

logger = logging.getLogger(__name__)

TRS_ROOT = "https://dockstore.org" if "TOIL_TRS_ROOT" not in os.environ else os.environ["TOIL_TRS_ROOT"]

def is_trs_workflow(workflow: str) -> bool:
    """
    Returns True if a workflow string smells like TRS.

    Detects Dockstore page URLs and strings that could be Dockstore TRS IDs.
    """

    return workflow.startswith(f"{TRS_ROOT}/workflows/") or workflow.startswith(f"{TRS_ROOT}/my-workflows/") or workflow.startswith("#workflow/")

def extract_trs_spec(workflow: str) -> str:
    """
    Parse a Dockstore workflow URL or TSR ID to a string that is definitely a TRS ID.
    """

    if workflow.startswith("#workflow/"):
        # Looks like a Dockstore TRS ID already.
        # TODO: Does Dockstore guartantee we can recognize its TRS IDs like this?
        logger.debug("Workflow %s is a TRS specifier already", workflow)
        trs_spec = workflow
    else:
        # We need to get the right TRS ID from the Dockstore URL
        parsed = urlparse(workflow)
        # TODO: We assume the Dockstore page URL structure and the TRS IDs are basically the same.
        page_path = unquote(parsed.path)
        if page_path.startswith("/workflows/"):
            trs_spec = "#workflow/" + page_path[len("/workflows/"):]
        elif page_path.startswith("/my-workflows/"):
            trs_spec = "#workflow/" + page_path[len("/my-workflows/"):]
        else:
            raise RuntimeError("Cannot parse Dockstore URL " + workflow)
        logger.debug("Translated %s to TRS: %s", workflow, trs_spec)

    return trs_spec

def parse_trs_spec(trs_spec: str) -> tuple[str, Optional[str]]:
    """
    Parse a TRS ID to workflow and optional version.
    """
    parts = trs_spec.split(':', 1)
    trs_workflow_id = parts[0]
    if len(parts) > 1:
        # The ID has the version we want after a colon
        trs_version = parts[1]
    else:
        # We don't know the version we want, we will have to pick one somehow.
        trs_version = None
    return trs_workflow_id, trs_version

def compose_trs_spec(trs_workflow_id: str, trs_version: str) -> str:
    """
    Compose a TRS ID from a workflow ID and version ID.
    """
    return f"{trs_workflow_id}:{trs_version}"

@retry(errors=[requests.exceptions.ConnectionError])
def find_workflow(workflow: str, supported_languages: Optional[set[str]] = None) -> tuple[str, str, str]:
    """
    Given a Dockstore URL or TRS identifier, get the root WDL or CWL URL for the workflow, along with the TRS workflow ID and version.

    Accepts inputs like:

        - https://dockstore.org/workflows/github.com/dockstore-testing/md5sum-checker:master?tab=info
        - #workflow/github.com/dockstore-testing/md5sum-checker

    Assumes the input is actually one of the supported formats. See is_trs_workflow().

    TODO: Needs to handle multi-workflow files if Dockstore can.

    :raises FileNotFoundError: if the workflow or version doesn't exist.
    :raises ValueError: if the version is not specified but cannot be
        automatically determined. 
    """

    if supported_languages is not None and len(supported_languages) == 0:
        raise ValueError("Set of supported languages must be nonempty if provided.")

    # Get the TRS id[:version] string from what might be a Dockstore URL
    trs_spec = extract_trs_spec(workflow)
    # Parse out workflow and possible version
    trs_workflow_id, trs_version = parse_trs_spec(trs_spec)

    logger.debug("TRS %s parses to workflow %s and version %s", trs_spec, trs_workflow_id, trs_version)

    # Fetch the main TRS document.
    # See e.g. https://dockstore.org/api/ga4gh/trs/v2/tools/%23workflow%2Fgithub.com%2Fdockstore-testing%2Fmd5sum-checker
    trs_workflow_url = f"{TRS_ROOT}/api/ga4gh/trs/v2/tools/{quote(trs_workflow_id, safe='')}"
    logger.debug("Get versions: %s", trs_workflow_url)
    trs_workflow_response = web_session.get(trs_workflow_url)
    if trs_workflow_response.status_code in (400, 404):
        # If the workflow ID isn't in Dockstore's accepted format (and also thus doesn't exist), we can get a 400
        raise FileNotFoundError(f"Workflow {trs_workflow_id} does not exist.")
    trs_workflow_response.raise_for_status()
    trs_workflow_document = trs_workflow_response.json()

    # Make a map from version to version info. We will need the
    # "descriptor_type" array to find eligible languages, and the "url" field
    # to get the version's base URL.
    workflow_versions: dict[str, dict[str, Any]] = {}

    # We also check which we actually know how to run
    eligible_workflow_versions: set[str] = set()

    for version_info in trs_workflow_document.get("versions", []):
        version_name: str = version_info["name"]
        workflow_versions[version_name] = version_info
        version_languages: list[str] = version_info["descriptor_type"]
        if supported_languages is not None:
            # Filter to versions that have a language we know
            has_supported_language = False
            for language in version_languages:
                if language in supported_languages:
                    # TODO: Also use "descriptor_type_version" dict to make
                    # sure we support all needed language versions to actually
                    # use this workflow version.
                    has_supported_language = True
                    continue
            if not has_supported_language:
                # Can't actually run this one.
                continue
        eligible_workflow_versions.add(version_name)

    # TODO: Dockstore has a concept of a "default version", but doesn't expose
    # it over TRS. To avoid defaulting to something that *isn't* the Dockstore
    # default version, we refuse to choose a version when there are multiple
    # possibilities.

    if trs_version is None and len(eligible_workflow_versions) == 1:
        # If there's just one version use that.
        trs_version = next(iter(eligible_workflow_versions))
        logger.debug("Defaulting to only eligible workflow version %s", trs_version)


    # If we don't like what we found we compose a useful error message.
    problems: list[str] = []
    problem_type: type[Exception] = RuntimeError
    if trs_version is None:
        problems.append(f"Workflow {workflow} does not specify a version")
        problem_type = ValueError
    elif trs_version not in workflow_versions:
        problems.append(f"Workflow version {trs_version} from {workflow} does not exist")
        problem_type = FileNotFoundError
    elif trs_version not in eligible_workflow_versions:
        message = f"Workflow version {trs_version} from {workflow} is not available"
        if supported_languages is not None:
            message += f" in any of: {', '.join(supported_languages)}"
        problems.append(message)
        problem_type = FileNotFoundError
    if len(problems) > 0:
        if len(eligible_workflow_versions) == 0:
            message = "No versions of the workflow are available"
            if supported_languages is not None:
                message += f" in any of: {', '.join(supported_languages)}"
            problems.append(message)
            problem_type = FileNotFoundError
        elif trs_version is None:
            problems.append(f"Add ':' and the name of a workflow version ({', '.join(eligible_workflow_versions)}) after '{trs_workflow_id}'")
        else:
            problems.append(f"Replace '{trs_version}' with one of ({', '.join(eligible_workflow_versions)})")
        raise problem_type("; ".join(problems))

    # Tell MyPy we now have a version, or we would have raised
    assert trs_version is not None

    # Select the language we will actually run
    chosen_version_languages: list[str] = workflow_versions[trs_version]["descriptor_type"]
    for candidate_language in chosen_version_languages:
        if supported_languages is None or candidate_language in supported_languages:
            language = candidate_language

    logger.debug("Going to use %s version %s in %s", trs_workflow_id, trs_version, language)

    return trs_workflow_id, trs_version, language
    
@retry(errors=[requests.exceptions.ConnectionError])
def fetch_workflow(trs_workflow_id: str, trs_version: str, language: str) -> str:
    """
    Returns a URL or local path to a workflow's primary descriptor file.

    The file will be in context with its required files so it can actually run.

    :raises FileNotFoundError: if the workflow or version doesn't exist.
    """

    # TODO: We should probably use HATEOAS and pull this from the worflow
    # document we probably already fetched but aren't passing.
    trs_version_url = f"{TRS_ROOT}/api/ga4gh/trs/v2/tools/{quote(trs_workflow_id, safe='')}/versions/{quote(trs_version, safe='')}"

    # Fetch the list of all the files
    trs_files_url = f"{trs_version_url}/{language}/files"
    logger.debug("Workflow files URL: %s", trs_files_url)
    trs_files_response = web_session.get(trs_files_url)
    if trs_files_response.status_code in (204, 400, 404):
        # We can get a 204 No Content response if the version doesn't exist.
        # That's successful, so we need to handle it specifically. See
        # <https://github.com/dockstore/dockstore/issues/6048>
        # We can also get a 400 if the workflow ID is not in Dockstore's expected format (3 slash-separated segments).
        raise FileNotFoundError(f"Workflow {trs_workflow_id} version {trs_version} in language {language} does not exist.")
    trs_files_response.raise_for_status()
    trs_files_document = trs_files_response.json()

    # Find the information we need to ID the primary descriptor file
    primary_descriptor_path: Optional[str] = None
    primary_descriptor_hash_algorithm: Optional[str] = None
    primary_descriptor_hash: Optional[str] = None
    for file_info in trs_files_document:
        if file_info["file_type"] == "PRIMARY_DESCRIPTOR":
            primary_descriptor_path = file_info["path"]
            primary_descriptor_hash_algorithm = file_info["checksum"]["type"]
            primary_descriptor_hash = file_info["checksum"]["checksum"]
            break
    if primary_descriptor_path is None or primary_descriptor_hash is None or primary_descriptor_hash_algorithm is None:
        raise RuntimeError(f"Could not find a primary descriptor file for workflow {trs_workflow_id} version {trs_version} in language {language}")
    primary_descriptor_basename = os.path.basename(primary_descriptor_path)

    # Work out how to compute the hash we are looking for. See
    # <https://github.com/ga4gh-discovery/ga4gh-checksum/blob/master/hash-alg.csv>
    # for the GA4GH names and <https://docs.python.org/3/library/hashlib.html>
    # for the Python names.
    #
    # TODO: We don't support the various truncated hash flavors or the other checksums not in hashlib.
    python_hash_name = primary_descriptor_hash_algorithm.replace("sha-", "sha").replace("blake2b-512", "blake2b").replace("-", "_")
    if python_hash_name not in hashlib.algorithms_available:
        raise RuntimeError(f"Primary descriptor is identified by a {primary_descriptor_hash_algorithm} hash but {python_hash_name} is not available in hashlib")

    # Figure out where to store the workflow. We don't want to deal with temp
    # dir cleanup since we don't want to run the whole workflow setup and
    # execution in a context manager. So we declare a cache.
    # Note that it's still not safe to symlink out of this cache since XDG
    # cache directories aren't guaranteed to be on shared storage.
    cache_base_dir = os.path.join(os.environ.get("XDG_CACHE_HOME", os.path.expanduser("~/.cache")), "toil/workflows")

    # Hash the workflow file list.
    hasher = hashlib.sha256()
    for file_info in sorted(trs_files_document, key=lambda rec: rec["path"]):
        hasher.update(file_info["path"].encode("utf-8"))
        hasher.update(b"\0")
        hasher.update(file_info["checksum"]["type"].encode("utf-8"))
        hasher.update(b"\0")
        hasher.update(file_info["checksum"]["checksum"].encode("utf-8"))
        hasher.update(b"\0")
    cache_workflow_dir = os.path.join(cache_base_dir, hasher.hexdigest())

    if os.path.exists(cache_workflow_dir):
        logger.debug("Workflow already cached at %s", cache_workflow_dir)
    else:
        # Need to download the workflow

        # Download the ZIP to a temporary file
        trs_zip_file_url = f"{trs_files_url}?format=zip"
        logger.debug("Workflow ZIP URL: %s", trs_zip_file_url)
        with tempfile.NamedTemporaryFile(suffix=".zip") as zip_file:
            # We want to stream the zip to a file, but when we do it with the Requests
            # file object like <https://stackoverflow.com/a/39217788> we don't get
            # Requests' decoding of gzip or deflate response encodings. Since this file
            # is already compressed the response compression can't help a lot anyway,
            # so we tell the server that we can't understand it.
            headers = {
                "Accept-Encoding": "identity",
                # Help Dockstore avoid serving ZIP with a JSON content type. See
                # <https://github.com/dockstore/dockstore/issues/6010>.
                "Accept": "application/zip"
            }
            # If we don't set stream=True, we can't actually read anything from the
            # raw stream, since Requests will have done it already.
            with web_session.get(trs_zip_file_url, headers=headers, stream=True) as response:
                response_content_length = response.headers.get("Content-Length")
                logger.debug("Server reports content length: %s", response_content_length)
                shutil.copyfileobj(response.raw, zip_file)
            zip_file.flush()

            logger.debug("Downloaded ZIP to %s", zip_file.name)

            # Unzip it to a directory next to where it will live
            os.makedirs(cache_base_dir, exist_ok=True)
            workflow_temp_dir = tempfile.mkdtemp(dir=cache_base_dir)
            with zipfile.ZipFile(zip_file.name, "r") as zip_ref:
                zip_ref.extractall(workflow_temp_dir)
            logger.debug("Extracted workflow ZIP to %s", workflow_temp_dir)

            # Try to atomically install into the cache
            try:
                os.rename(workflow_temp_dir, cache_workflow_dir)
                logger.debug("Moved workflow to %s", cache_workflow_dir)
            except OSError:
                # Collision. Someone else installed the workflow before we could.
                robust_rmtree(workflow_temp_dir)
                logger.debug("Workflow cached at %s by someone else while we were donwloading it", cache_workflow_dir)

    # Hunt throught he directory for a file with the right basename and hash
    found_path: Optional[str] = None
    for containing_dir, subdirectories, files in os.walk(cache_workflow_dir):
        for filename in files:
            if filename == primary_descriptor_basename:
                # This could be it. Open the file off disk and hash it with the right algorithm.
                file_path = os.path.join(containing_dir, filename)
                file_hash = file_digest(open(file_path, "rb"), python_hash_name).hexdigest()
                if file_hash == primary_descriptor_hash:
                    # This looks like the right file
                    logger.debug("Found candidate primary descriptor %s", file_path)
                    if found_path is not None:
                        # But there are multiple instances of it so we can't know which to run.
                        # TODO: Find out the right path from Dockstore somehow!
                        raise RuntimeError(f"Workflow contains multiple files named {primary_descriptor_basename} with {python_hash_name} hash {file_hash}: {found_path} and {file_path}")
                    # This is the first file with the right name and hash
                    found_path = file_path
                else:
                    logger.debug("Rejected %s because its %s hash %s is not %s", file_path, python_hash_name, file_hash, primary_descriptor_hash)
    if found_path is None:
        # We couldn't find the promised primary descriptor
        raise RuntimeError(f"Could not find a {primary_descriptor_basename} with {primary_descriptor_hash_algorithm} hash {primary_descriptor_hash} for workflow {trs_workflow_id} version {trs_version} in language {language}")

    return found_path

def resolve_workflow(workflow: str, supported_languages: Optional[set[str]] = None) -> tuple[str, Optional[str]]:
    """
    Find the real workflow URL or filename from a command line argument.

    Transform a workflow URL or path that might actually be a Dockstore page
    URL or TRS specifier to an actual URL or path to a workflow document, and
    optional TRS specifier.

    Accepts inputs like

        - https://dockstore.org/workflows/github.com/dockstore-testing/md5sum-checker:master?tab=info
        - #workflow/github.com/dockstore-testing/md5sum-checker
        - ./local.cwl
        - https://example.com/~myuser/workflow/main.cwl

    :raises FileNotFoundError: if the workflow or version should be in Dockstore but doesn't seem to exist.
    """

    if is_trs_workflow(workflow):
        # Ask TRS host where to find TRS-looking things
        trs_workflow_id, trs_version, language = find_workflow(workflow, supported_languages)
        resolved = fetch_workflow(trs_workflow_id, trs_version, language)
        logger.info("Resolved TRS workflow %s to %s", workflow, resolved)
        return resolved, compose_trs_spec(trs_workflow_id, trs_version)
    else:
        # Pass other things through.
        # TODO: Find out if they have TRS names.
        return workflow, None













from typing import Sequence

from _typeshed import Incomplete

from .annotations import annotationType

class Topic: ...

class TopicNameError(ValueError):
    def __init__(self, name: str, msg: str) -> None: ...

class TopicDefnError(RuntimeError):
    def __init__(self, topicNameTuple: Sequence[str]) -> None: ...

class MessageDataSpecError(RuntimeError):
    def __init__(self, msg: str, args: Sequence[str]) -> None: ...

class ExcHandlerError(RuntimeError):
    badExcListenerID: Incomplete
    exc: Incomplete
    def __init__(
        self, badExcListenerID: str, topicObj: Topic, origExc: Exception = ...
    ) -> None: ...

class UnrecognizedSourceFormatError(ValueError):
    def __init__(self) -> None: ...
